"""
This module contains a collection of classes and methods for normalizing features. Also included is connection with
scikit-learn methods. See http://scikit-learn.org/stable/modules/classes.html#module-sklearn.preprocessing for more info.
"""

from functools import wraps

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import MinMaxScaler, Binarizer, StandardScaler, MaxAbsScaler, Normalizer, \
QuantileTransformer, RobustScaler, OneHotEncoder


from mastml.legos import util_legos

# TODO: add all sklearn preprocessors
# http://scikit-learn.org/stable/modules/classes.html#module-sklearn.preprocessing

def dataframify(transform):
    """
    Method which is a decorator transforms output of scikit-learn feature normalizers from array to dataframe.
    Enables preservation of column names.

    Args:

        transform: (function), a scikit-learn feature selector that has a transform method

    Returns:

        new_transform: (function), an amended version of the transform method that returns a dataframe

    """

    @wraps(transform)
    def new_transform(self, df):
        arr = transform(self, df.values)
        return pd.DataFrame(arr, columns=df.columns, index=df.index)
    return new_transform

class MeanStdevScaler(BaseEstimator, TransformerMixin):
    """
    Class designed to normalize input data to a specified mean and standard deviation

    Args:

        mean: (int/float), specified normalized mean of the data

        stdev: (int/float), specified normalized standard deviation of the data

    Methods:

        fit: Obtains initial mean and stdev of data

            Args:

                df: (dataframe), dataframe of values to be normalized

            Returns:

                (self, the object instance)

        transform: Normalizes the data to new mean and stdev values

            Args:

                df: (dataframe), dataframe of values to be normalized

            Returns:

                (dataframe), dataframe containing re-normalized data and any data that wasn't normalized

        inverse_transform: Un-normalizes the data to the old mean and stdev values

            Args:

                df: (dataframe), dataframe of values to be un-normalized

            Returns:

                (dataframe), dataframe containing un-normalized data and any data that wasn't normalized

    """

    def __init__(self, features=None, mean=0, stdev=1):
        self.features = features
        self.mean = mean
        self.stdev = stdev

    def fit(self, df, y=None):
        if self.features is None:
            self.features = df.columns
        self.old_mean  = df[self.features].values.mean()
        self.old_stdev = df[self.features].values.std()
        return self

    def transform(self, df):
        array = df[self.features].values
        array = ((array - self.old_mean) / self.old_stdev) * self.stdev + self.mean
        same = df.drop(columns=self.features)
        changed = pd.DataFrame(array, columns=self.features, index=df.index) # don't forget index!!
        return pd.concat([same, changed], axis=1)

    def inverse_transform(self, df):
        array = df[self.features].values
        array = ((array - self.mean) / self.stdev) * self.old_stdev + self.old_mean
        same = df.drop(columns=self.features)
        changed = pd.DataFrame(array, columns=self.features, index=df.index)
        return pd.concat([same, changed], axis=1)

MinMaxScaler.transform = dataframify(MinMaxScaler.transform)
Binarizer.transform = dataframify(Binarizer.transform)
StandardScaler.transform = dataframify(StandardScaler.transform)
MaxAbsScaler.transform = dataframify(MaxAbsScaler.transform)
Normalizer.transform = dataframify(Normalizer.transform)
QuantileTransformer.transform = dataframify(QuantileTransformer.transform)
RobustScaler.transform = dataframify(RobustScaler.transform)
OneHotEncoder.transform = dataframify(OneHotEncoder.transform)

name_to_constructor = {
    'Binarizer': Binarizer,
    'MeanStdevScaler': MeanStdevScaler,
    'MinMaxScaler': MinMaxScaler,
    'MaxAbsScaler': MaxAbsScaler,
    'Normalizer': Normalizer,
    'QuantileTransformer': QuantileTransformer,
    'RobustScaler': RobustScaler,
    'OneHotEncoder': OneHotEncoder,
    'DoNothing': util_legos.DoNothing,
    'StandardScaler': StandardScaler
}



"""
This module provides a name_to_constructor dict for all models/estimators in scikit-learn, plus a couple test models and
error handling functions
"""

import warnings
import inspect

import sklearn.base
import sklearn.utils.testing
from sklearn.externals import joblib
import numpy as np

# Sometimes xgboost is hard to install so make it optional
try:
    import xgboost as xgb
except:
    pass

import keras
from keras.models import model_from_json
from keras.models import load_model
from keras.models import Sequential

#from . import keras_models
from mastml import utils

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    name_to_constructor = dict(sklearn.utils.testing.all_estimators())

class AlwaysFive(sklearn.base.RegressorMixin):
    """
    Class used as a test model that always predicts a value of 5.

    Args:

        constant: (int), the value to predict. Always 5 by default

    Methods:

        fit: Just passes through to maintain scikit-learn structure

        predict: Provides predicted model values based on X features

            Args:

                X: (numpy array), array of X features

            Returns:

                (numpy array), prediction array where all values are equal to constant

    """
    def __init__(self, constant = 5):
        self.five = constant

    def fit(self, X, y, groups=None):
        return self

    def predict(self, X):
        return np.array([self.five for _ in range(len(X))])

class RandomGuesser(sklearn.base.RegressorMixin):
    """
    Class used as a test model that always predicts random values for y data.

    Args:

        None

    Methods:

        fit: Constructs possible predicted values based on y data

            Args:

                y: (numpy array), array of y data

        predict: Provides predicted model values based on X features

            Args:

                X: (numpy array), array of X features

            Returns:

                (numpy array), prediction array where all values are random selections of y data

    """
    def __init__(self):
        pass

    def fit(self, X, y, groups=None):
        self.possible_answers = y
        return self

    def predict(self, X):
        return np.random.choice(self.possible_answers, size=X.shape[0])

class KerasRegressor():
    def __init__(self, conf_dict):
        self.conf_dict = conf_dict
        self.model = self.build_model()

    def build_model(self):
        model_vals = self.conf_dict
        model = Sequential()

        for layer_dict, layer_val in model_vals.items():
            if (layer_dict != 'FitParams'):
                layer_type = layer_val.get('layer_type')
                layer_name_asstr = layer_type
                if layer_name_asstr == 'Dense':
                    neuron_num = int(layer_val.get('neuron_num'))
                    if (layer_dict == 'Layer1'):
                        input_dim = int(layer_val.get('input_dim'))
                    kernel_initializer = layer_val.get('kernel_initializer')
                    activation = layer_val.get('activation')
                elif layer_name_asstr == 'Dropout':
                    rate = float(layer_val.get('rate'))
                for layer_name, cls in inspect.getmembers(keras.layers, inspect.isclass):
                    layer_type = getattr(keras.layers, layer_name_asstr)  # (neuron_num)

            else:
                if layer_val.get('rate'):
                    self.rate = float(layer_val.get('rate'))
                if layer_val.get('epochs'):
                    self.epochs = int(layer_val.get('epochs'))
                else:
                    self.epochs = 1
                if layer_val.get('batch_size'):
                    self.batch_size = int(layer_val.get('batch_size'))
                else:
                    self.batch_size = None
                if layer_val.get('loss'):
                    self.loss = str(layer_val.get('loss'))
                else:
                    self.loss = 'mean_squared_error'
                if layer_val.get('optimizer'):
                    self.optimizer = str(layer_val.get('optimizer'))
                else:
                    self.optimizer = 'adam'
                if layer_val.get('metrics'):
                    self.metrics = layer_val.get('metrics').split(',')
                else:
                    self.metrics = ['mae']
                if layer_val.get('verbose'):
                    self.verbose = str(layer_val.get('verbose'))
                else:
                    self.verbose = 0
                if layer_val.get('shuffle'):
                    self.shuffle = bool(layer_val.get('shuffle'))
                else:
                    self.shuffle = True
                if layer_val.get('validation_split'):
                    self.validation_split = float(layer_val.get('validation_split'))
                else:
                    self.validation_split = 0.0
                continue

            if (layer_dict == 'Layer1'):
                model.add(layer_type(neuron_num, input_dim=input_dim, kernel_initializer=kernel_initializer,
                                     activation=activation))

            else:
                if layer_name_asstr == 'Dense':
                    model.add(layer_type(neuron_num, kernel_initializer=kernel_initializer, activation=activation))
                if layer_name_asstr == 'Dropout':
                    model.add(layer_type(rate=rate))

        return model

    def fit(self, X, Y):
        # Need to rebuild and re-compile model at every fit instance so don't have information of weights from other fits
        self.model = self.build_model()
        self.model.compile(loss=self.loss, optimizer=self.optimizer, metrics=self.metrics)
        return self.model.fit(X, Y, epochs=self.epochs, batch_size=self.batch_size, verbose=self.verbose,
                              validation_split=self.validation_split, shuffle=self.shuffle)

    def predict(self, X):
        return self.model.predict(X)

    def summary(self):
        return self.model.summary()

class ModelImport():
    """
    Class used to import pickled models from previous machine learning fits

    Args:

        model_path (str): string designating the path to load the saved .pkl model file

    Methods:

        fit: Does nothing, present for compatibility purposes

            Args:

                X: Nonetype

                y: Nonetype

                groups: Nonetype

        predict: Provides predicted model values based on X features

            Args:

                X: (numpy array), array of X features

            Returns:

                (numpy array), prediction array using imported model

    """
    def __init__(self, model_path):
        self.model_path = model_path
        self.model = joblib.load(self.model_path)

    def fit(self, X=None, y=None, groups=None):
        """ Only here for compatibility """
        return

    def predict(self, X):
        return self.model.predict(X)

# Optional to have xgboost working
try:
    custom_models = {
        'AlwaysFive': AlwaysFive,
        'RandomGuesser': RandomGuesser,
        'ModelImport': ModelImport,
        'XGBRegressor': xgb.XGBRegressor,
        'XGBClassifier': xgb.XGBClassifier,
        'KerasRegressor': KerasRegressor
        #'DNNClassifier': keras_models.DNNClassifier
    }
except NameError:
    custom_models = {
        'AlwaysFive': AlwaysFive,
        'RandomGuesser': RandomGuesser,
        'ModelImport': ModelImport,
        'KerasRegressor': KerasRegressor
        # 'DNNClassifier': keras_models.DNNClassifier
    }
name_to_constructor.update(custom_models)

def find_model(model_name):
    """
    Method used to check model names conform to scikit-learn model/estimator names

    Args:

        model_name: (str), the name of a model/estimator

    Returns:

        (str), the scikit-learn model name or raises InvalidModel error

    """
    try:
        return name_to_constructor[model_name]
    except KeyError:
        raise utils.InvalidModel(f"Model '{model_name}' does not exist in scikit-learn.")

def check_models_mixed(model_names):
    """
    Method used to check whether the user has mixed regression and classification tasks

    Args:

        model_names: (list), list containing names of models/estimators

    Returns:

        (bool), whether or not a classifier was found, or raises exception if both regression and classification models present.

    """

    found_classifier = found_regressor = False
    for name in model_names:
        if name in custom_models: continue
        class1 = find_model(name)
        if issubclass(class1, sklearn.base.ClassifierMixin):
            found_classifier = True
        elif issubclass(class1, sklearn.base.RegressorMixin):
            found_regressor = True
        else:
            raise Exception(f"Model '{name}' is neither a classifier nor a regressor")

    if found_classifier and found_regressor:
        raise Exception("Both classifiers and regressor models have been included")

    return found_classifier




"""
This module contains a collection of classes for debugging and control flow
"""
import logging

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

log = logging.getLogger('mastml')

class DataFrameFeatureUnion(BaseEstimator, TransformerMixin):
    """
    Class for unioning dataframe generators (sklearn.pipeline.FeatureUnion always puts out arrays)

    Args:

        transforms: (list), list of scikit-learn functions, i.e. objects with a .fit or .transform method

    Methods:

        fit: Applies the .fit method for each transform

        Args:

            X: (numpy array), array of X features

        transform: Transforms the output of the scikit-learn transformer into a dataframe

        Args:

            X: (numpy array), array of X features

        Returns:

            (dataframe), concatenated dataframe after all scikit-learn transforms have been completed

    """

    def __init__(self, transforms):
        self.transforms = transforms

    def fit(self, X, y=None):
        for transform in self.transforms:
            transform.fit(X, y)
        return self

    def transform(self, X):
        dataframes = [transform.transform(X) for transform in self.transforms]
        return pd.concat(dataframes, axis=1)

class DoNothing(BaseEstimator, TransformerMixin):
    """
    Class for having a "null" transform where the output is the same as the input. Needed by MAST-ML as a placeholder if
    certain workflow aspects are not performed.

    Args:

        None

    Methods:

        fit: does nothing, just returns object instance. Needed to maintain same structure as scikit-learn classes

        Args:

            X: (numpy array), array of X features

        transform: passes the input back out, in this case the array of X features

        Args:

            X: (numpy array), array of X features

        Returns:

            X: (numpy array), array of X features

    """

    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X



"""
This module contains constructors for different model score metrics. Most model metrics are obtained from scikit-learn,
while others are custom variations.

The full list of score functions in scikit-learn can be found at: http://scikit-learn.org/stable/modules/model_evaluation.html
"""

import numpy as np
import sklearn.feature_selection as fs
import sklearn.metrics as sm
from sklearn.linear_model import LinearRegression # for r2_score_noint

classification_metrics = {
    'accuracy':           (True, sm.accuracy_score),
    'f1_binary':          (True, lambda yt, yp: sm.f1_score(yt, yp, average='binary')),
    'f1_macro':           (True, lambda yt, yp: sm.f1_score(yt, yp, average='macro')),
    'f1_micro':           (True, lambda yt, yp: sm.f1_score(yt, yp, average='micro')),
    'f1_samples':         (True, lambda yt, yp: sm.f1_score(yt, yp, average='samples')),
    'f1_weighted':        (True, lambda yt, yp: sm.f1_score(yt, yp, average='weighted')),
    'log_loss':           (False, sm.log_loss),
    'precision_binary':   (True, lambda yt, yp: sm.precision_score(yt, yp, average='binary')),
    'precision_macro':    (True, lambda yt, yp: sm.precision_score(yt, yp, average='macro')),
    'precision_micro':    (True, lambda yt, yp: sm.precision_score(yt, yp, average='micro')),
    'precision_samples':  (True, lambda yt, yp: sm.precision_score(yt, yp, average='samples')),
    'precision_weighted': (True, lambda yt, yp: sm.precision_score(yt, yp, average='weighted')),
    'recall_binary':      (True, lambda yt, yp: sm.recall_score(yt, yp, average='binary')),
    'recall_macro':       (True, lambda yt, yp: sm.recall_score(yt, yp, average='macro')),
    'recall_micro':       (True, lambda yt, yp: sm.recall_score(yt, yp, average='micro')),
    'recall_samples':     (True, lambda yt, yp: sm.recall_score(yt, yp, average='samples')),
    'recall_weighted':    (True, lambda yt, yp: sm.recall_score(yt, yp, average='weighted')),
    'roc_auc':            (True, sm.roc_auc_score),
}

regression_metrics = {
    'explained_variance':     (True, sm.explained_variance_score),
    'mean_absolute_error':    (False, sm.mean_absolute_error),
    'mean_squared_error':     (False, sm.mean_squared_error),
    'mean_squared_log_error': (False, sm.mean_squared_log_error),
    'median_absolute_error':  (False, sm.median_absolute_error),
    'R2': (True, sm.r2_score)
}

def r2_score_noint(y_true, y_pred):
    """
    Method that calculates the R^2 value without fitting the y-intercept

    Args:
        y_true: (numpy array), array of true y data values
        y_pred: (numpy array), array of predicted y data values

    Returns:
        (float): score of R^2 with no y-intercept

    """
    lr = LinearRegression(fit_intercept=False)
    y_true = np.array(y_true).reshape(-1,1) # turn it from an n-vector to nx1-matrix
    lr.fit(y_true, y_pred)
    return lr.score(y_true, y_pred)
regression_metrics['R2_noint'] = (True, r2_score_noint)

def r2_score_fitted(y_true, y_pred):
    """
    Method that calculates the R^2 value

    Args:
        y_true: (numpy array), array of true y data values
        y_pred: (numpy array), array of predicted y data values

    Returns:
        (float): score of R^2

    """
    lr = LinearRegression(fit_intercept=True)
    y_true = np.array(y_true).reshape(-1,1) # turn it from an n-vector to nx1-matrix
    lr.fit(y_true, y_pred)
    return lr.score(y_true, y_pred)
regression_metrics['R2_fitted'] = (True, r2_score_fitted)

def root_mean_squared_error(y_true, y_pred):
    """
    Method that calculates the root mean squared error (RMSE)

    Args:
        y_true: (numpy array), array of true y data values
        y_pred: (numpy array), array of predicted y data values

    Returns:
        (float): score of RMSE

    """
    return sm.mean_squared_error(y_true, y_pred)**0.5
regression_metrics['root_mean_squared_error'] = (False, root_mean_squared_error)

# TODO: consider two rmse/stdev metrics: one that uses stdev of full data set, and one that uses stdev of data per split
def rmse_over_stdev(y_true, y_pred, train_y=None):
    """
    Method that calculates the root mean squared error (RMSE) of a set of data, divided by the standard deviation of
    the training data set.

    Args:
        y_true: (numpy array), array of true y data values
        y_pred: (numpy array), array of predicted y data values
        train_y: (numpy array), array of training y data values

    Returns:
        (float): score of RMSE divided by standard deviation of training data

    """
    if train_y is not None:
        stdev = np.std(train_y)
    else:
        stdev = np.std(y_true)
    rmse = root_mean_squared_error(y_true, y_pred)
    return rmse / stdev
regression_metrics['rmse_over_stdev'] = (False, rmse_over_stdev)

def adjusted_r2_score(y_true, y_pred, n_features=None):
    """
    Method that calculates the adjusted R^2 value

    Args:
        y_true: (numpy array), array of true y data values
        y_pred: (numpy array), array of predicted y data values
        n_features: (int), number of features used in the fit

    Returns:
        (float): score of adjusted R^2

    """
    r2 = r2_score(y_true, y_pred)
    # n is sample size
    n = len(y_true)
    # p is number of features
    p = n_features
    try:
        r2_score_adj = 1 - (((1-r2)*(n-1))/(n-p-1))
    except:
        # No n_features given, just output NaN
        r2_score_adj = 'NaN'
    return r2_score_adj
regression_metrics['R2_adjusted'] = (True, adjusted_r2_score)

classification_score_funcs = {
    'chi2': fs.chi2, # Compute chi-squared stats between each non-negative feature and class.
    'f_classif': fs.f_classif, # Compute the ANOVA F-value for the provided sample.
    'mutual_info_classif': fs.mutual_info_classif, # Estimate mutual information for a discrete target variable.
}

regression_score_funcs = {
    'f_regression': fs.f_regression, # Univariate linear regression tests.
    'mutual_info_regression': fs.mutual_info_regression, # Estimate mutual information for a continuous target variable.
}

nice_names = {
    # classification:
    'accuracy': 'Accuracy',
    'f1_binary': '$F_1$',
    'f1_macro': 'f1_macro',
    'f1_micro': 'f1_micro',
    'f1_samples': 'f1_samples',
    'f1_weighted': 'f1_weighted',
    'log_loss': 'log_loss',
    'precision_binary': 'Precision',
    'precision_macro': 'prec_macro',
    'precision_micro': 'prec_micro',
    'precision_samples': 'prec_samples',
    'precision_weighted': 'prec_weighted',
    'recall_binary': 'Recall',
    'recall_macro': 'rcl_macro',
    'recall_micro': 'rcl_micro',
    'recall_samples': 'rcl_samples',
    'recall_weighted': 'rcl_weighted',
    'roc_auc': 'ROC_AUC',
    # regression:
    'explained_variance': 'expl_var',
    'mean_absolute_error': 'MAE',
    'mean_squared_error': 'MSE',
    'mean_squared_log_error': 'MSLE',
    'median_absolute_error': 'MedAE',
    'root_mean_squared_error': 'RMSE',
    'rmse_over_stdev': r'RMSE/$\sigma_y$',
    'R2': '$R^2$',
    'R2_noint': '$R^2_{noint}$',
    'R2_adjusted': '$R^2_{adjusted}$'
}

def check_and_fetch_names(metric_names, is_classification):
    """
    Method that checks whether chosen metrics to evaluate models are appropriate for user-specified models (e.g.
    classification vs. regression models)

    Args:
        metric_names: (numpy array), array of true y data values
        is_classification: (bool), whether the task is a classification task

    Returns:
        functions (dict): dict containing the appropriate metric objects (e.g. classification vs. regression metrics)

    """
    " Ensures all metrics are appropriate for task "
    task = 'classification' if is_classification else 'regression'
    metrics_dict = classification_metrics if is_classification else regression_metrics
    functions = {}
    for name in metric_names:
        if name not in metrics_dict:
            raise Exception(f"Metric '{name}' is not supported for {task}.\n"
                            f"Valid metrics for {task}: {list(metrics_dict.keys())}")
        functions[name] = metrics_dict[name]
    return functions



"""
This module contains a collection of classes and methods for selecting features, and interfaces with scikit-learn feature
selectors. More information on scikit-learn feature selectors is available at:

http://scikit-learn.org/stable/modules/classes.html#module-sklearn.feature_selection
"""

from functools import wraps
import warnings
import numpy as np
from mastml.metrics import root_mean_squared_error

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.decomposition import PCA
import sklearn.feature_selection as fs
from mlxtend.feature_selection import SequentialFeatureSelector
import os, logging

log = logging.getLogger('mastml')

from mastml.legos import util_legos

def dataframify_selector(transform):
    """
    Method which transforms output of scikit-learn feature selectors from array to dataframe. Enables preservation of column names.

    Args:

        transform: (function), a scikit-learn feature selector that has a transform method

    Returns:

        new_transform: (function), an amended version of the transform method that returns a dataframe

    """

    @wraps(transform)
    def new_transform(self, df):
        if isinstance(df, pd.DataFrame):
            return df[df.columns[self.get_support(indices=True)]]
        else: # just in case you try to use it with an array ;)
            return df
    return new_transform

def dataframify_new_column_names(transform, name):
    """
    Method which transforms output of scikit-learn feature selectors to dataframe, and adds column names

    Args:

        transform: (function), a scikit-learn feature selector that has a transform method

        name: (str), name of the feature selector

    Returns:

        new_transform: (function), an amended version of the transform method that returns a dataframe

    """

    def new_transform(self, df):
        arr = transform(self, df.values)
        labels = [name+str(i) for i in range(arr.shape[1])]
        return pd.DataFrame(arr, columns=labels)
    return new_transform

def fitify_just_use_values(fit):
    """
    Method which enables a feature selector fit method to operate on dataframes

    Args:

        fit: (function), a scikit-learn feature selector object with a fit method

    Returns:

        new_fit: (function), an amended version of the fit method that uses dataframes as input

    """

    def new_fit(self, X_df, y_df):
        return fit(self, X_df.values, y_df.values)
    return new_fit

score_func_selectors = {
    'GenericUnivariateSelect': fs.GenericUnivariateSelect, # Univariate feature selector with configurable strategy.
    'SelectFdr': fs.SelectFdr, # Filter: Select the p-values for an estimated false discovery rate
    'SelectFpr': fs.SelectFpr, # Filter: Select the pvalues below alpha based on a FPR test.
    'SelectFwe': fs.SelectFwe, # Filter: Select the p-values corresponding to Family-wise error rate
    'SelectKBest': fs.SelectKBest, # Select features according to the k highest scores.
    'SelectPercentile': fs.SelectPercentile, # Select features according to a percentile of the highest scores.
}

model_selectors = { # feature selectors which take a model instance as first parameter
    'RFE': fs.RFE, # Feature ranking with recursive feature elimination.
    'RFECV': fs.RFECV, # Feature ranking with recursive feature elimination and cross-validated selection of the best number of features.
    'SelectFromModel': fs.SelectFromModel, # Meta-transformer for selecting features based on importance weights.
}

other_selectors = {
    'VarianceThreshold': fs.VarianceThreshold, # Feature selector that removes all low-variance features.
}

# Union together the above dicts for the primary export:
name_to_constructor = dict(**score_func_selectors, **model_selectors, **other_selectors)

# Modify all sklearn transform methods to return dataframes:
for constructor in name_to_constructor.values():
    constructor.old_transform = constructor.transform
    constructor.transform = dataframify_selector(constructor.transform)


class EnsembleModelFeatureSelector(object):

    def __init__(self, estimator, k_features):
        self.estimator = estimator
        self.k_features = k_features
        # Check that a correct model was passed in
        self._check_model()
        self.selected_features = list()

    def _check_model(self):
        if self.estimator.__class__.__name__ not in ['RandomForestRegressor', 'ExtraTreesRegressor', 'GradientBoostingRegressor']:
            raise ValueError('Models used in EnsembleModelFeatureSelector must be one of RandomForestRegressor, ExtraTreesRegressor, GradientBoostingRegressor')
        return

    def fit(self, X, y=None):
        feature_importances = self.estimator.fit(X, y).feature_importances_
        feature_importance_dict = dict()
        for col, f in zip(X.columns.tolist(), feature_importances):
            feature_importance_dict[col] = f
        feature_importances_sorted = sorted(((f, col) for col, f in feature_importance_dict.items()), reverse=True)
        sorted_features_list = [f[1] for f in feature_importances_sorted]
        self.selected_features = sorted_features_list[0:self.k_features]
        return self

    def transform(self, X):
        df = X[self.selected_features]
        return df


class MASTMLFeatureSelector(object):
    """
    Class custom-written for MAST-ML to conduct forward selection of features with flexible model and cv scheme

    Args:

        estimator: (scikit-learn model/estimator object), a scikit-learn model/estimator

        n_features_to_select: (int), the number of features to select

        cv: (scikit-learn cross-validation object), a scikit-learn cross-validation object

        manually_selected_features: (list), a list of features manually set by the user. The feature selector will first
        start from this list of features and sequentially add features until n_features_to_select is met.

    Methods:

        fit: performs feature selection

            Args:

                X: (dataframe), dataframe of X features

                y: (dataframe), dataframe of y data

                Xgroups: (dataframe), dataframe of group labels

            Returns:

                None

        transform: performs the transform to generate output of only selected features

            Args:

                X: (dataframe), dataframe of X features

            Returns:

                dataframe: (dataframe), dataframe of selected X features

    """

    def __init__(self, estimator, n_features_to_select, cv, manually_selected_features=list()):
        self.estimator = estimator
        self.n_features_to_select = n_features_to_select
        self.cv = cv
        self.manually_selected_features = manually_selected_features
        self.selected_feature_names = self.manually_selected_features

    def fit(self, X, y, savepath, Xgroups=None):
        if Xgroups.shape[0] == 0:
            xgroups = np.zeros(len(y))
            Xgroups = pd.DataFrame(xgroups)

        selected_feature_avg_rmses = list()
        selected_feature_std_rmses = list()
        basic_forward_selection_dict = dict()
        num_features_selected = 0
        x_features = X.columns.tolist()
        if self.n_features_to_select >= len(x_features):
            self.n_features_to_select = len(x_features)
        while num_features_selected < self.n_features_to_select:
            log.info('On number of features selected')
            log.info(str(num_features_selected))

            # Catch pandas warnings here
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                ranked_features = self._rank_features(X=X, y=y, groups=Xgroups)
                top_feature_name, top_feature_avg_rmse, top_feature_std_rmse = self._choose_top_feature(ranked_features=ranked_features)

            self.selected_feature_names.append(top_feature_name)
            if len(self.selected_feature_names) > 0:
                log.info('selected features')
                log.info(self.selected_feature_names)
            selected_feature_avg_rmses.append(top_feature_avg_rmse)
            selected_feature_std_rmses.append(top_feature_std_rmse)

            basic_forward_selection_dict[str(num_features_selected)] = dict()
            basic_forward_selection_dict[str(num_features_selected)][
                'Number of features selected'] = num_features_selected + 1
            basic_forward_selection_dict[str(num_features_selected)][
                'Top feature added this iteration'] = top_feature_name
            basic_forward_selection_dict[str(num_features_selected)][
                'Avg RMSE using top features'] = top_feature_avg_rmse
            basic_forward_selection_dict[str(num_features_selected)][
                'Stdev RMSE using top features'] = top_feature_std_rmse
            # Save for every loop of selecting features
            pd.DataFrame(basic_forward_selection_dict).to_csv(os.path.join(savepath,'MASTMLFeatureSelector_data_feature_'+str(num_features_selected)+'.csv'))
            num_features_selected += 1
        basic_forward_selection_dict[str(self.n_features_to_select - 1)][
            'Full feature set Names'] = self.selected_feature_names
        basic_forward_selection_dict[str(self.n_features_to_select - 1)][
            'Full feature set Avg RMSEs'] = selected_feature_avg_rmses
        basic_forward_selection_dict[str(self.n_features_to_select - 1)][
            'Full feature set Stdev RMSEs'] = selected_feature_std_rmses
        #self._plot_featureselected_learningcurve(selected_feature_avg_rmses=selected_feature_avg_rmses,
        #                                         selected_feature_std_rmses=selected_feature_std_rmses)

        return self

    def transform(self, X):
        dataframe = self._get_featureselected_dataframe(X=X, selected_feature_names=self.selected_feature_names)
        return dataframe

    def _rank_features(self, X, y, groups):
        y = np.array(y).reshape(-1, 1)
        ranked_features = dict()
        trains_metrics = list()
        tests_metrics = list()
        if groups is not None:
            groups = groups.iloc[:,0].tolist()
        for col in X.columns:
            if col not in self.selected_feature_names:
                X_ = X.loc[:, self.selected_feature_names]
                X__ = X.loc[:, col]
                X_ = np.array(pd.concat([X_, X__], axis=1))

                for trains, tests in self.cv.split(X_, y, groups):
                    self.estimator.fit(X_[trains], y[trains])
                    predict_tests = self.estimator.predict(X_[tests])
                    tests_metrics.append(root_mean_squared_error(y[tests], predict_tests))
                avg_rmse = np.mean(tests_metrics)

                std_rmse = np.std(tests_metrics)
                ranked_features[col] = {"avg_rmse": avg_rmse, "std_rmse": std_rmse}
        return ranked_features

    def _choose_top_feature(self, ranked_features):
        feature_names = list()
        feature_avg_rmses = list()
        feature_std_rmses = list()
        feature_names_sorted = list()
        feature_std_rmses_sorted = list()
        # Make dict of ranked features into list for sorting
        for k, v in ranked_features.items():
            feature_names.append(k)
            for kk, vv in v.items():
                if kk == 'avg_rmse':
                    feature_avg_rmses.append(vv)
                if kk == 'std_rmse':
                    feature_std_rmses.append(vv)

        # Sort feature lists so RMSE goes from min to max
        feature_avg_rmses_sorted = sorted(feature_avg_rmses)
        for feature_avg_rmse in feature_avg_rmses_sorted:
            for k, v in ranked_features.items():
                if v['avg_rmse'] == feature_avg_rmse:
                    feature_names_sorted.append(k)
                    feature_std_rmses_sorted.append(v['std_rmse'])

        top_feature_name = feature_names_sorted[0]
        top_feature_avg_rmse = feature_avg_rmses_sorted[0]
        top_feature_std_rmse = feature_std_rmses_sorted[0]

        return top_feature_name, top_feature_avg_rmse, top_feature_std_rmse

    def _get_featureselected_dataframe(self, X, selected_feature_names):
        # Return dataframe containing only selected features
        X_selected = X.loc[:, selected_feature_names]
        return X_selected


# Include Principal Component Analysis
PCA.transform = dataframify_new_column_names(PCA.transform, 'pca_')

# Include Sequential Forward Selector
SequentialFeatureSelector.transform = dataframify_new_column_names(SequentialFeatureSelector.transform, 'sfs_')
SequentialFeatureSelector.fit = fitify_just_use_values(SequentialFeatureSelector.fit)
model_selectors['SequentialFeatureSelector'] = SequentialFeatureSelector
name_to_constructor['SequentialFeatureSelector'] = SequentialFeatureSelector

# Custom selectors don't need to be dataframified
name_to_constructor.update({
    #'PassThrough': PassThrough,
    'DoNothing': util_legos.DoNothing,
    'PCA': PCA,
    'SequentialFeatureSelector': SequentialFeatureSelector,
    'MASTMLFeatureSelector' : MASTMLFeatureSelector,
    'EnsembleModelFeatureSelector': EnsembleModelFeatureSelector
})


import os

_ROOT = os.path.abspath(os.path.dirname(__file__))
def get_data(path):
    return os.path.join(_ROOT, path)

paths = list()
for folders, subfolders, files in os.walk(_ROOT):
    paths.append(os.path.join(_ROOT, files))

for path in paths:
    print(get_data(path=path))


"""
Module for generating an HTML file, called index.html, which contains an overview of the key data and plots from a
MAST-ML run. Images of cross-validation parity plots, data histograms, data statistics, and links to the relevant files
are all provided.
"""

import os
from os.path import join, relpath # because it's used so much
from time import gmtime, strftime
import logging
from dominate import document
from dominate.tags import *

log = logging.getLogger('mastml')

def make_html(outdir):
    """
    Method used to create the main index.html file

    Args:

        outdir: (str), user-specified output path which designates where all results of MAST-ML run are written

    Returns:

        None

    """

    with document(title='MASTML') as doc:
        # title and date
        h1('MAterial Science Tools - Machine Learning')
        h4(strftime("%Y-%m-%d %H:%M:%S", gmtime()))

        # link to error log
        #if errors_present:
        #    p('You have errors! check ', make_link(error_log))

        combos = list()
        link_sections = list()
        #favorites = dict()

        for root, dirs, files in os.walk(outdir):
            # find a folder that contains split_ folder.
            # For example, results/StandardScaler/SelectKBest/LinearRegression/KFold
            for d in dirs:
                if d.startswith('split_0'):
                    combos.append(root)

            # extract links to important csvs and conf
            for f in files:
                csv_whitelist = [
                    'clusters.csv', 'generated_features.csv',
                    'generated_features_no_constant_columns.csv', 'grouped.csv',
                    'input_data_statistics.csv', 'normalized.csv', 'selected.csv', ]
                ext = os.path.splitext(f)[1]
                if f in csv_whitelist or ext in ['.conf', '.log']:
                    link_sections.append(join(root, f))
                    #simple_section(join(root, f), outdir)

        h1('Files')
        for path in link_sections:
            simple_section(path, outdir)

        h1('Plots')

        # show all the images
        for f in os.listdir(outdir):
            if f.endswith('.png'):
                make_image(f, f)

        for combo in combos:
            # come up with a good section title
            path = os.path.normpath(relpath(combo, outdir))
            paths = path.split(os.sep)
            title = " - ".join(paths)
            h2(title)

            # find the best worst overlay
            for fname in os.listdir(combo):
                if fname.endswith('.png'):
                    h3(os.path.splitext(fname)[0]) # probably best_worst overlay
                    make_image(relpath(join(combo, fname), outdir), fname)
                    br()

            # find the split_0 split_1 etc bs stuff
            for fname in os.listdir(combo):
                if fname.startswith('split_'):
                    show_combo(join(combo, fname), outdir)

    with open(join(outdir, 'index.html'), 'w') as f:
        f.write(doc.render())

    log.info('wrote ' + join(outdir, 'index.html'))

def show_combo(combo_dir, outdir):
    """
    Method used to collect combinations of data analysis (e.g. parity plots of train and test data in a CV split) and
    required file paths and display them in the output index.html file.

    Args:

        combo_dir: (str), path containing the relevant data to combine as output in the index.html file

        outdir: (str), user-specified output path which designates where all results of MAST-ML run are written

    Returns:

        None

    """

    # collect test image, train image, and other file links
    links = list()
    train_images = list()
    test_images = list()
    for f in os.listdir(combo_dir):
        if is_train_image(f):
            train_images.append(join(combo_dir, f))
        elif is_test_image(f):
            test_images.append(join(combo_dir, f))
        else:
            links.append(join(combo_dir, f))

    # have a header for split_0 split_1 etc
    h2(combo_dir.split(os.sep)[-1])

    # loop separately so we can control order
    for train_image, test_image in zip(sorted(train_images), sorted(test_images)):
        make_image(relpath(train_image, outdir), 'train')
        make_image(relpath(test_image, outdir), 'test')
        br();br()

    h3('links')
    for l in links:
        make_link(relpath(l, outdir))
        span('  ')

def simple_section(filepath, outdir):
    """
    Method used to create a section name for a particular analysis combination that will be displayed in the index.html file.

    Args:

        filepath: (str), path containing the relevant data to combine as output in the index.html file

        outdir: (str), user-specified output path which designates where all results of MAST-ML run are written

    Returns:

        None

    """

    " Create a section for a combo "
    path = os.path.normpath(relpath(filepath, outdir))
    paths = path.split(os.sep)
    title = " - ".join(paths)
    a(b(title))
    make_link(relpath(filepath, outdir))
    br()

def make_link(href):
    """
    Method used to generate a link to a particular file created from a MAST-ML run. The link will be displayed next to the
    appropriate data or image in the index.html file

    Args:

        href: (str), filename to generate link for

    Returns:

        (dominate.tags html_tag object), hyperlink to filename

    """

    " Make a link where text is filename of href "
    return a(os.path.basename(href), href=href, style='padding-left: 15px;')

def make_image(src, title=None):
    """
    Method used to generate and show an image of a fixed width. The image will be displayed in the appropriate
    section of the index.html file

    Args:

        src: (str), source path of the image to be displayed

        title: (str), title for the image

    Returns:

        None

    """
    " Show an image in fixed width "
    d = div(style='display:inline-block;', _class='photo')
    if title:
        d += h4(title)
        #d += p(a(title))
    d += img(src=src, height='200')

def is_train_image(path):
    """
    Method used to assess whether an image is for training data

    Args:

        path: (str), source path of the image to be displayed

    Returns:

        (bool), True if path is an image (.png) and is for training data (has 'train' in path)

    """
    basename = os.path.basename(path)
    return os.path.splitext(basename)[1] == '.png' and 'train' in basename

def is_test_image(path):
    """
    Method used to assess whether an image is for testing data

    Args:

        path: (str), source path of the image to be displayed

    Returns:

        (bool), True if path is an image (.png) and is for testing data (has 'test' in path)

    """
    basename = os.path.basename(path)
    return os.path.splitext(basename)[1] == '.png' and 'test' in basename



"""
Main MAST-ML module responsible for executing the workflow of a MAST-ML run
"""

import argparse
import inspect
import os
import shutil
import logging
import warnings
import re
import json
from datetime import datetime
from collections import OrderedDict
from os.path import join # We use join tons
from functools import reduce
from contextlib import redirect_stdout

import numpy as np
import pandas as pd
from sklearn.externals import joblib
from sklearn.exceptions import UndefinedMetricWarning
from sklearn.model_selection import LeaveOneGroupOut
from sklearn.metrics import make_scorer

from mastml import conf_parser, data_loader, html_helper, plot_helper, utils, learning_curve, data_cleaner, metrics
from mastml.legos import (data_splitters, feature_generators, feature_normalizers,
                    feature_selectors, model_finder, util_legos, randomizers, hyper_opt)
from mastml.legos import clusterers as legos_clusterers
from mastml.legos import model_hosting

log = logging.getLogger('mastml')

def main(conf_path, data_path, outdir=join(os.getcwd(), 'results_mastml_run'), verbosity=0):
    """
    This method is responsible for setting up the initial stage of the MAST-ML run, such as parsing input directories to
    designate where data will be imported and results saved to, as well as creation of the MAST-ML run log.

    Args:

        conf_path: (str), the path supplied by the user which contains the input configuration file

        data_path: (str), the path supplied by the user which contains the input data file (as CSV or XLSX)

        outdir: (str), the path supplied by the user which determines where the output results are saved to

        verbosity: (int), the verbosity level of the MAST-ML log, which determines the amount of information written to the log.

    Returns:

        outdir: (str), the path supplied by the user which determines where the output results are saved to (needed by other calls in MAST-ML)

    """

    conf_path, data_path, outdir = check_paths(conf_path, data_path, outdir)

    utils.activate_logging(outdir, (str(conf_path), str(data_path), outdir), verbosity=verbosity)

    if verbosity >= 1:
        warnings.simplefilter('error') # turn warnings into errors
    elif verbosity <= -1:
        warnings.simplefilter('ignore') # ignore warnings

    try:
        mastml_run(conf_path, data_path, outdir)
    except utils.MastError as e:
        # catch user errors, log and print, but don't raise and show them that nasty stack
        log.error(str(e))
    except Exception as e:
        # catch the error, save it to file, then raise it back up
        log.error('A runtime exception has occured, please go to '
                      'https://github.com/uw-cmg/MAST-ML/issues and post your issue.')
        log.exception(e)
        raise e
    return outdir # so a calling program can know where we actually saved it

def mastml_run(conf_path, data_path, outdir):
    """
    This method is responsible for conducting the main MAST-ML run workflow

    Args:

        conf_path: (str), the path supplied by the user which contains the input configuration file

        data_path: (str), the path supplied by the user which contains the input data file (as CSV or XLSX)

        outdir: (str), the path supplied by the user which determines where the output results are saved to

    Returns:

        None

    """


    " Runs operations specifed in conf_path on data_path and puts results in outdir "

    # Copy the original input files to the output directory for easy reference
    log.info("Copying input files to output directory...")
    if type(conf_path) is str:
        shutil.copy2(conf_path, outdir)
    elif type(conf_path) is dict:
        with open(join(outdir, 'conf_file.conf'), 'w') as f:
            json.dump(conf_path, f)

    if type(data_path) is str:
        shutil.copy2(data_path, outdir)
    elif type(data_path) is type(pd.DataFrame()):
        data_path.to_excel(join(outdir, 'input_data.xlsx'), index=False)
        data_path = join(outdir, 'input_data.xlsx')

    # Load in and parse the configuration and data files:
    if type(conf_path) is str:
        conf = conf_parser.parse_conf_file(conf_path, from_dict=False)
    elif type(conf_path) is dict:
        conf = conf_parser.parse_conf_file(conf_path, from_dict=True)
    else:
        raise TypeError('Your conf_path must either be a path string to a .conf file or a dict directly containing the config info')

    MiscSettings = conf['MiscSettings']
    is_classification = conf['is_classification']
    # The df is used by feature generators, clusterers, and grouping_column to 
    # create more features for x.
    # X is model input, y is target feature for model
    df, X, X_noinput, X_grouped, y = data_loader.load_data(data_path,
                                         conf['GeneralSetup']['input_features'],
                                         conf['GeneralSetup']['input_target'],
                                         conf['GeneralSetup']['input_grouping'],
                                         conf['GeneralSetup']['input_other'])
    if not conf['GeneralSetup']['input_grouping']:
        X_grouped = pd.DataFrame()

    # Perform data cleaning here
    dc = conf['DataCleaning']
    if 'cleaning_method' not in dc.keys():
        log.warning("You have chosen not to specify a method of data_cleaning in the input file. By default, any feature entries "
                    "containing NaN will result in removal of the feature and any target data entries containing NaN will "
                    "result in removal of that target data point.")
        dc['cleaning_method'] = 'remove'

    if X.shape[1] == 0:
        # There are no X feature vectors specified, so can't clean data
        log.warning("There are no X feature vectors imported from the data file. Therefore, data cleaning cannot be performed.")
    else:
        # Always scan the input data and flag potential outliers
        data_cleaner.flag_outliers(df=df, conf_not_input_features=conf['GeneralSetup']['input_other'],
                                   savepath=outdir,
                                   n_stdevs=3)
        if dc['cleaning_method'] == 'remove':
            df, nan_indices = data_cleaner.remove(df, axis=1)
            X, nan_indices = data_cleaner.remove(X, axis=1)
            X_noinput, nan_indices = data_cleaner.remove(X_noinput, axis=1)
            X_grouped, nan_indices = data_cleaner.remove(X_grouped, axis=1)
        elif dc['cleaning_method'] == 'imputation':
            log.warning("You have selected data cleaning with Imputation. Note that imputation will not resolve missing target data. "
                        "It is recommended to remove missing target data")
            if 'imputation_strategy' not in dc.keys():
                log.warning("You have chosen to perform data imputation but have not selected an imputation strategy. By default, "
                            "the mean will be used as the imputation strategy")
                dc['imputation_strategy'] = 'mean'
            df = data_cleaner.imputation(df, dc['imputation_strategy'], X_noinput.columns)
            X = data_cleaner.imputation(X, dc['imputation_strategy'])
        elif dc['cleaning_method'] == 'ppca':
            log.warning("You have selected data cleaning with PPCA. Note that PPCA will not work to estimate missing target values, "
                        "at least a 2D matrix is needed. It is recommended you remove missing target data")
            df = data_cleaner.ppca(df, X_noinput.columns)
            X = data_cleaner.ppca(X)
        else:
            log.error("You have specified an invalid data cleaning method. Choose from: remove, imputation, or ppca")
            exit()

        # Check if any y target data values are missing or NaN
        shape_before = y.shape
        y, nan_indices = data_cleaner.remove(y, axis=0)
        shape_after = y.shape
        if shape_after != shape_before:
            log.info(
                'Warning: some y target data rows were automatically removed because they were either empty or contained '
                '"NaN" entries. MAST-ML will continue your run with this modified data set.')
            # Need to modify df, X, etc. to remove same rows as were removed from y target data
            df = df.drop(labels=nan_indices, axis=0)
            X = X.drop(labels=nan_indices, axis=0)
            X_noinput = X_noinput.drop(labels=nan_indices, axis=0)
            if X_grouped.shape[0] > 0:
                X_grouped = X_grouped.drop(labels=nan_indices, axis=0)

    # randomly shuffles y values if randomizer is on
    if conf['GeneralSetup']['randomizer'] is True:
        log.warning("Randomizer is enabled, so target feature will be shuffled,"
                 " and results should be null for a given model")
        y = randomizers.Randomizer().fit().transform(df=y)

    """
        # get parameters out for 'validation_column'
    is_validation = 'validation_columns' in conf['GeneralSetup']
    if is_validation:
        if type(conf['GeneralSetup']['validation_columns']) is list:
            validation_column_names = list(conf['GeneralSetup']['validation_columns'])
        elif type(conf['GeneralSetup']['validation_columns']) is str:
            validation_column_names = list()
            validation_column_names.append(conf['GeneralSetup']['validation_columns'][:])
        validation_columns = dict()
        for validation_column_name in validation_column_names:
            validation_columns[validation_column_name] = df[validation_column_name]
        validation_columns = pd.DataFrame(validation_columns)
        validation_X = list()
        validation_y = list()

        # TODO make this block its own function
        for validation_column_name in validation_column_names:
            # X_, y_ = _exclude_validation(X, validation_columns[validation_column_name]), _exclude_validation(y, validation_columns[validation_column_name])
            validation_X.append(pd.DataFrame(_exclude_validation(X, validation_columns[validation_column_name])))
            validation_y.append(pd.DataFrame(_exclude_validation(y, validation_columns[validation_column_name])))
        idxy_list = list()
        for i, _ in enumerate(validation_y):
            idxy_list.append(validation_y[i].index)
        # Get intersection of indices between all prediction columns
        intersection = reduce(np.intersect1d, (i for i in idxy_list))
        X_novalidation = X.iloc[intersection]
        y_novalidation = y.iloc[intersection]
        X_grouped_novalidation = X_grouped.iloc[intersection]
    else:
        X_novalidation = X
        y_novalidation = y
        X_grouped_novalidation = X_grouped
    """


    if conf['MiscSettings']['plot_target_histogram']:
        # First, save input data stats to csv
        y.describe().to_csv(join(outdir, 'input_data_statistics.csv'))
        plot_helper.plot_target_histogram(y, join(outdir, 'target_histogram.png'), label=y.name)

    # Get the appropriate collection of metrics:
    metrics_dict = conf['GeneralSetup']['metrics']

    # Extract columns that some splitter need to do grouped splitting using 'grouping_column'
    # special argument
    splitter_to_group_names = _extract_grouping_column_names(conf['DataSplits'])
    log.debug('splitter_to_group_names:\n' + str(splitter_to_group_names))

    # Instantiate models first so we can snatch them and pass them into feature selectors
    models = _instantiate(conf['Models'],
                          model_finder.name_to_constructor,
                          'model')

    models = _snatch_models(models, conf['FeatureSelection'])

    # Instantiate all the sections of the conf file:
    generators  = _instantiate(conf['FeatureGeneration'],
                               feature_generators.name_to_constructor,
                               'featuregenerator')
    clusterers  = _instantiate(conf['Clustering'],
                               legos_clusterers.name_to_constructor,
                               'clusterer')
    normalizers = _instantiate(conf['FeatureNormalization'],
                               feature_normalizers.name_to_constructor,
                               'featurenormalizer')
    splitters   = _instantiate(conf['DataSplits'],
                               data_splitters.name_to_constructor,
                               'datasplit')

    def snatch_model_cv_and_scoring_for_learning_curve(models):
        models = OrderedDict(models)
        if conf['LearningCurve']:
            # Get model
            name = conf['LearningCurve']['estimator']
            conf['LearningCurve']['estimator'] = models[name]
            del models[name]
            # Get cv
            name = conf['LearningCurve']['cv']
            splitter_count = 0
            for splitter in splitters:
                if name in splitter:
                    conf['LearningCurve']['cv'] = splitter[1]
                    break
                else:
                    splitter_count += 1
            del splitters[splitter_count]
        return models

    models = snatch_model_cv_and_scoring_for_learning_curve(models=models)

    models = _snatch_keras_model(models, conf['Models'])
    original_models = models

    # Need to specially snatch the GPR model if it is in models list because it contains special kernel object
    models = _snatch_gpr_model(models, conf['Models'])
    original_models = models

    # Need to snatch models and CV objects for Hyperparam Opt
    hyperopt_params = _snatch_models_cv_for_hyperopt(conf, models, splitters)

    hyperopts = _instantiate(hyperopt_params,
                             hyper_opt.name_to_constructor,
                             'hyperopt')

    hyperopts = OrderedDict(hyperopts)
    hyperopts = list(hyperopts.items())

    # Snatch splitter for use in feature selection, particularly RFECV
    splitters = OrderedDict(splitters)  # for easier modification
    _snatch_splitters(splitters, conf['FeatureSelection'])
    splitters = list(splitters.items())

    selectors   = _instantiate(conf['FeatureSelection'],
                               feature_selectors.name_to_constructor,
                               'featureselector', X_grouped=np.array(X_grouped).reshape(-1, ), X_indices=np.array(X.index.tolist()).reshape(-1, 1))

    log.debug(f'generators: \n{generators}')
    log.debug(f'clusterers: \n{clusterers}')
    log.debug(f'normalizers: \n{normalizers}')
    log.debug(f'hyperopts: \n{hyperopts}')
    log.debug(f'selectors: \n{selectors}')
    log.debug(f'splitters: \n{splitters}')

    # TODO make this block its own function, and change naming from is_validation to is_test here and throughout. Just symantic annoyance.
    # get parameters out for 'validation_column'
    is_validation = 'input_testdata' in conf['GeneralSetup']
    if is_validation:
        if type(conf['GeneralSetup']['input_testdata']) is list:
            validation_column_names = list(conf['GeneralSetup']['input_testdata'])
        elif type(conf['GeneralSetup']['input_testdata']) is str:
            validation_column_names = list()
            validation_column_names.append(conf['GeneralSetup']['input_testdata'][:])
        validation_columns = dict()
        for validation_column_name in validation_column_names:
            validation_columns[validation_column_name] = df[validation_column_name]
        validation_columns = pd.DataFrame(validation_columns)
        validation_X = list()
        validation_y = list()

    def do_all_combos(X, y, df):
        log.info(f"There are {len(normalizers)} feature normalizers, {len(hyperopts)} hyperparameter optimizers, "
                 f"{len(selectors)} feature selectors, {len(models)} models, and {len(splitters)} splitters.")

        def generate_features():
            log.info("Doing feature generation...")
            dataframes = [instance.fit_transform(df, y) for _, instance in generators]
            dataframe = pd.concat(dataframes, 1)
            log.info("Saving generated data to csv...")
            log.debug(f'generated cols: {dataframe.columns}')
            filename = join(outdir, "generated_features.csv")
            pd.concat([dataframe, X_noinput, y], 1).to_csv(filename, index=False)
            return dataframe
        generated_df = generate_features()

        def remove_constants():
            dataframe = _remove_constant_features(generated_df)
            log.info("Saving generated data without constant columns to csv...")
            filename = join(outdir, "generated_features_no_constant_columns.csv")
            pd.concat([dataframe, X_noinput, y], 1).to_csv(filename, index=False)
            return dataframe
        generated_df = remove_constants()

        # add in generated features
        X = pd.concat([X, generated_df], axis=1)
        # add in generated features to full dataframe
        df = pd.concat([df, generated_df], axis=1)
        # Check size of X; if there are no feature columns then throw error
        if X.shape[1] == 0:
            raise utils.InvalidValue('No feature vectors were found in the dataframe. Please either use feature generation methods'
                               'or specify input_features in the input file.')

        # remove repeat columns (keep the first one)
        def remove_repeats(X):
            repeated_columns = X.loc[:, X.columns.duplicated()].columns
            if not repeated_columns.empty:
                log.warning(f"Throwing away {len(repeated_columns)} because they are repeats.")
                log.debug(f"Throwing away columns because they are repeats: {repeated_columns}")
                X = X.loc[:,~X.columns.duplicated()]
            return X
        X = remove_repeats(X)

        # TODO make this block its own function
        # get parameters out for 'validation_column'
        if is_validation:
            for validation_column_name in validation_column_names:
                # X_, y_ = _exclude_validation(X, validation_columns[validation_column_name]), _exclude_validation(y, validation_columns[validation_column_name])
                validation_X.append(pd.DataFrame(_exclude_validation(X, validation_columns[validation_column_name])))
                validation_y.append(pd.DataFrame(_exclude_validation(y, validation_columns[validation_column_name])))
            idxy_list = list()
            for i, _ in enumerate(validation_y):
                idxy_list.append(validation_y[i].index)
            # Get intersection of indices between all prediction columns
            intersection = reduce(np.intersect1d, (i for i in idxy_list))
            X_novalidation = X.iloc[intersection]
            y_novalidation = y.iloc[intersection]
            if conf['GeneralSetup']['input_grouping']:
                X_grouped_novalidation = X_grouped.iloc[intersection]
            else:
                X_grouped_novalidation = pd.DataFrame()
        else:
            X_novalidation = X
            y_novalidation = y
            if conf['GeneralSetup']['input_grouping']:
                X_grouped_novalidation = X_grouped
            else:
                X_grouped_novalidation = pd.DataFrame()

        def make_clustered_df():
            log.info("Doing clustering...")
            clustered_df = pd.DataFrame()
            for name, instance in clusterers:
                clustered_df[name] = instance.fit_predict(X, y)
            return clustered_df
        clustered_df = make_clustered_df() # Each column is a clustering algorithm

        def make_feature_vs_target_plots():
            if clustered_df.empty:
                for column in X: # plot y against each x column
                    filename = f'{column}_vs_target_scatter.png'
                    plot_helper.plot_scatter(X[column], y, join(outdir, filename),
                                             xlabel=column, groups=None, label=y.name)
            else:
                for name in clustered_df.columns: # for each cluster, plot y against each x column
                    for column in X:
                        filename = f'{column}_vs_target_by_{name}_scatter.png'
                        plot_helper.plot_scatter(X[column], y, join(outdir, filename),
                                                clustered_df[name], xlabel=column, label=y.name)
        if MiscSettings['plot_each_feature_vs_target']:
            make_feature_vs_target_plots()

        log.info("Saving clustered data to csv...")
        # Add new cluster info to X df
        if not clustered_df.empty:
            X = pd.concat([X, clustered_df], axis=1)
        pd.concat([X, y], 1).to_csv(join(outdir, "clusters.csv"), index=False)

        def make_normalizer_selector_dataframe_triples(models):
            triples = []
            nonlocal y, y_novalidation
            for normalizer_name, normalizer_instance in normalizers:

                # Run feature normalization
                log.info(f"Running normalizer {normalizer_name} ...")
                normalizer_instance_y = normalizer_instance
                normalizer_instance_y_novalidation = normalizer_instance

                # HERE- try to address issue with normalizing non-validation part of dataset
                normalizer = normalizer_instance.fit(X_novalidation, y)
                X_normalized = normalizer.transform(X)
                X_novalidation_normalized = normalizer.transform(X_novalidation)

                if conf['MiscSettings']['normalize_target_feature'] is True:
                    yreshape = pd.DataFrame(np.array(y).reshape(-1, 1))
                    y_novalidation_new = pd.DataFrame(np.array(y_novalidation).reshape(-1, 1))
                    y_normalized = normalizer_instance_y.fit_transform(yreshape, yreshape)
                    y_novalidation_normalized = normalizer_instance_y_novalidation.fit_transform(y_novalidation_new, y_novalidation_new)
                    y_normalized.columns = [conf['GeneralSetup']['input_target']]
                    y_novalidation_normalized.columns = [conf['GeneralSetup']['input_target']]
                    y = pd.Series(np.squeeze(y_normalized), name=conf['GeneralSetup']['input_target'])
                    y_novalidation = pd.Series(np.squeeze(y_novalidation_normalized), name=conf['GeneralSetup']['input_target'])
                else:
                    normalizer_instance_y = None
                log.info("Saving normalized data to csv...")
                dirname = join(outdir, normalizer_name)
                os.mkdir(dirname)
                pd.concat([X_normalized, X_noinput, y], 1).to_csv(join(dirname, "normalized.csv"), index=False)

                # Save off the normalizer as .pkl for future import
                joblib.dump(normalizer, join(dirname, str(normalizer.__class__.__name__) + ".pkl"))

                # HERE- find data twins


                # Put learning curve here??
                if conf['LearningCurve']:
                    learning_curve_estimator = conf['LearningCurve']['estimator']
                    learning_curve_scoring = conf['LearningCurve']['scoring']
                    n_features_to_select = int(conf['LearningCurve']['n_features_to_select'])
                    learning_curve_cv = conf['LearningCurve']['cv']
                    try:
                        selector_name = conf['LearningCurve']['selector_name']
                    except KeyError:
                        selector_name = None

                    # Get score name from scoring object
                    scoring_name = learning_curve_scoring._score_func.__name__
                    scoring_name_nice = ''
                    for s in scoring_name.split('_'):
                        scoring_name_nice += s + ' '
                    # Do sample learning curve
                    train_sizes, train_mean, test_mean, train_stdev, test_stdev = learning_curve.sample_learning_curve(X=X_novalidation_normalized, y=y_novalidation,
                                                            estimator=learning_curve_estimator, cv=learning_curve_cv,
                                                            scoring=learning_curve_scoring,
                                                            Xgroups=X_grouped_novalidation)
                    plot_helper.plot_learning_curve(train_sizes, train_mean, test_mean, train_stdev, test_stdev,
                                                    scoring_name_nice, 'sample_learning_curve',
                                                    join(dirname, f'data_learning_curve'))
                    # Do feature learning curve
                    train_sizes, train_mean, test_mean, train_stdev, test_stdev = learning_curve.feature_learning_curve(X=X_novalidation_normalized, y=y_novalidation,
                                                            estimator=learning_curve_estimator, cv=learning_curve_cv,
                                                            scoring=learning_curve_scoring, selector_name=selector_name,
                                                            savepath=dirname,
                                                            n_features_to_select=n_features_to_select,
                                                            Xgroups=X_grouped_novalidation)
                    plot_helper.plot_learning_curve(train_sizes, train_mean, test_mean, train_stdev, test_stdev,
                                                    scoring_name_nice, 'feature_learning_curve',
                                                    join(dirname, f'feature_learning_curve'))



                log.info("Running selectors...")

                # Run feature selection
                for selector_name, selector_instance in selectors:
                    log.info(f"    Running selector {selector_name} ...")
                    dirname = join(outdir, normalizer_name, selector_name)
                    os.mkdir(dirname)
                    # NOTE: Changed from .fit_transform to .fit.transform
                    # because PCA.fit_transform doesn't call PCA.transform
                    if selector_instance.__class__.__name__ == 'MASTMLFeatureSelector':
                        dirname = join(outdir, normalizer_name)
                        X_selected = selector_instance.fit(X_novalidation_normalized, y_novalidation, dirname, X_grouped_novalidation).transform(X_novalidation_normalized)
                    elif selector_instance.__class__.__name__ == 'SequentialFeatureSelector':
                        X_selected = selector_instance.fit(X_novalidation_normalized, y_novalidation).transform(X_novalidation_normalized)
                        # Need to reset indices in case have test data, otherwise df.equals won't properly find column names
                        X_novalidation_normalized_reset = X_novalidation_normalized.reset_index()
                        # SFS renames the columns. Need to replace the column names with correct feature names.
                        feature_name_dict = dict()
                        for feature in X_selected.columns.tolist():
                            for realfeature in X_novalidation_normalized.columns.tolist():
                                if X_novalidation_normalized_reset[realfeature].equals(X_selected[feature]):
                                    feature_name_dict[feature] = realfeature
                        X_selected.rename(columns= feature_name_dict, inplace=True)
                    elif selector_instance.__class__.__name__ == 'PearsonSelector':
                        X_selected = selector_instance.fit(X=X_novalidation_normalized, savepath=dirname, y=y_novalidation).transform(X_novalidation_normalized)
                    else:
                        X_selected = selector_instance.fit(X_novalidation_normalized, y_novalidation).transform(X_novalidation_normalized)
                    features_selected = X_selected.columns.tolist()
                    # Need to do this instead of taking X_selected directly because otherwise won't concatenate correctly with test data values, which are
                    # left out of the feature selection process.
                    X_selected = X_normalized[features_selected]
                    log.info("    Saving selected features to csv...")

                    pd.concat([X_selected, X_noinput, y], 1).to_csv(join(dirname, "selected.csv"), index=False)
                    #TODO: fix this naming convention
                    triples.append((normalizer_name, normalizer_instance_y, selector_name, X_selected))

                    # Run Hyperparam optimization, update model list with optimized model(s)
                    for hyperopt_name, hyperopt_instance in hyperopts:
                        try:
                            log.info(f"    Running hyperopt {hyperopt_name} ...")
                            log.info(f"    Saving optimized hyperparams and data to csv...")
                            dirname = join(outdir, normalizer_name, selector_name, hyperopt_name)
                            os.mkdir(dirname)
                            estimator_name = hyperopt_instance._estimator_name
                            best_estimator = hyperopt_instance.fit(X_selected, y, savepath=os.path.join(dirname, str(estimator_name)+'.csv'))

                            new_name = estimator_name + '_' + str(normalizer_name) + '_' + str(selector_name) + '_' + str(hyperopt_name)
                            new_model = best_estimator
                            models[new_name] = new_model

                            # Update models list with new hyperparams
                            #for model in models:
                            #    # model[0] is name, model[1] is instance
                            #    # Check that this particular model long_name had its hyperparams optimized
                            #    for name in hyperopt_params.keys():
                            #        if model[0] in name[:]:
                            #            model[0] = model[0]+'_nonoptimized'
                            #            #model[1] = best_estimator
                            #            # Need to update model as new model name and instance to handle multiple
                            #            # selector/optimization/fitting path
                            #            new_name = model[0]+'_'+str(normalizer_name)+'_'+str(selector_name)+'_'+str(hyperopt_name)
                            #            new_model = best_estimator
                            #            models[new_name] = new_model

                        except:
                            raise utils.InvalidValue

            return triples
        normalizer_selector_dataframe_triples = make_normalizer_selector_dataframe_triples(models=models)

        ## DataSplits (cross-product)
        ## Collect grouping columns, splitter_to_groupmes is a dict of splitter name to grouping col
        log.debug("Finding splitter-required columns in data...")
        def make_splittername_splitlist_pairs():
            # exclude the testing_only rows from use in splits
            if is_validation:
                validation_X = list()
                validation_y = list()
                for validation_column_name in validation_column_names:
                    #X_, y_ = _exclude_validation(X, validation_columns[validation_column_name]), _exclude_validation(y, validation_columns[validation_column_name])
                    validation_X.append(pd.DataFrame(_exclude_validation(X, validation_columns[validation_column_name])))
                    validation_y.append(pd.DataFrame(_exclude_validation(y, validation_columns[validation_column_name])))
                idxy_list = list()
                for i, _ in enumerate(validation_y):
                    idxy_list.append(validation_y[i].index)
                # Get intersection of indices between all prediction columns
                intersection = reduce(np.intersect1d, (i for i in idxy_list))
                X_ = X.iloc[intersection]
                y_ = y.iloc[intersection]
            else:
                X_, y_ = X, y

            pairs = []

            def fix_index(array):
                return X_.index.values[array] 

            def proper_index(splits):
                """ For example, if X's indexs are [1,4,6] and you split 
                [ [[0],[1,2]], [[1],[0,2]] ] then we would get 
                [ [[1],[4,6]], [[4],[1,6]] ] 
                Needed only for valdation row stuff.
                """
                return tuple(tuple(fix_index(part) for part in split) for split in splits)


            # Collect all the grouping columns, `None` if not needed
            splitter_to_group_column = dict()
            splitter_to_group_column_no_validation = dict()
            for name, instance in splitters:
                # if this splitter depends on grouping
                if name in splitter_to_group_names: 
                    col = splitter_to_group_names[name]
                    log.debug(f"    Finding {col} for {name}...")
                    # Locate the grouping column among all dataframes
                    for df_ in [clustered_df, df, X_]:
                        if col in df_.columns:
                            # FOund it!
                            # Get groups for plotting first
                            splitter_to_group_column[name] = df_[col].values
                            if is_validation:
                                _df_list = list()
                                if df_ is not clustered_df:
                                    # exclude for df_ so that rows match up in splitter
                                    for validation_column_name in validation_column_names:
                                        df_ = _exclude_validation(df_, validation_columns[validation_column_name])
                                        _df_list.append(df_)
                                elif df_ is clustered_df:
                                    # merge the cluster data df_ to full df
                                    df[col] = df_
                                    for validation_column_name in validation_column_names:
                                        df_ = _exclude_validation(df, validation_columns[validation_column_name])
                                        _df_list.append(df_)

                                # Get df_ based on index intersection between all df's in _df_list
                                idxy_list = list()
                                for i, _ in enumerate(_df_list):
                                    idxy_list.append(_df_list[i].index)
                                # Get intersection of indices between all prediction columns
                                intersection = reduce(np.intersect1d, (i for i in idxy_list))
                                df_ = df.iloc[intersection]

                            # and use the no-validation one for the split
                            grouping_data = df_[col].values
                            split = proper_index(instance.split(X_, y_, grouping_data))
                            pairs.append((name, split))
                            break
                    # If we didn't find that column anywhere, raise
                    else:
                        raise utils.MissingColumnError(f'DataSplit {name} needs column {col}, which '
                                                       f'was neither generated nor given by input')

                # If we don't need grouping column
                else: 
                    splitter_to_group_column[name] = None
                    split = proper_index(instance.split(X_, y_))
                    pairs.append((name, split))

            return pairs, splitter_to_group_column
        splittername_splitlist_pairs, splitter_to_group_column = make_splittername_splitlist_pairs()

        log.info("Fitting models to splits...")

        def do_models_splits(models, original_models):
            models = list(models.items())
            original_models = list(original_models.items())
            original_model_names = [model[0] for model in original_models]
            all_results = []
            for normalizer_name, normalizer_instance, selector_name, X in normalizer_selector_dataframe_triples:
                subdir = join(outdir, normalizer_name, selector_name)

                if MiscSettings['plot_each_feature_vs_target']:
                    #if selector_name == 'DoNothing': continue
                    # for each selector/normalizer, plot y against each x column
                    for column in X:
                        filename = f'{column}_vs_target.png'
                        plot_helper.plot_scatter(X[column], y, join(subdir, filename),
                                                 xlabel=column, label=y.name)
                for model_name, model_instance in models:
                    #Here, add logic to only run original models and respective models from hyperparam opt
                    do_split = False
                    if model_name in original_model_names:
                        do_split = True
                    elif (normalizer_name in model_name) and (selector_name in model_name):
                        do_split = True
                    if do_split == True:
                        for splitter_name, trains_tests in splittername_splitlist_pairs:
                            grouping_data = splitter_to_group_column[splitter_name]
                            subdir = join(normalizer_name, selector_name, model_name, splitter_name)
                            log.info(f"    Running splits for {subdir}")
                            subsubdir = join(outdir, subdir)
                            os.makedirs(subsubdir)
                            # NOTE: do_one_splitter is a big old function, does lots
                            runs = do_one_splitter(X, y, model_instance, subsubdir, trains_tests, grouping_data, normalizer_instance)
                            all_results.extend(runs)
            return all_results

        return do_models_splits(models, original_models)

    def do_one_splitter(X, y, model, main_path, trains_tests, grouping_data, normalizer_instance):

        def one_fit(split_num, train_indices, test_indices, normalizer_instance):

            log.info(f"        Doing split number {split_num}")
            train_X, train_y = X.loc[train_indices], y.loc[train_indices]
            test_X,  test_y  = X.loc[test_indices],  y.loc[test_indices]

            # split up groups into train and test as well
            if grouping_data is not None:
                train_groups, test_groups = grouping_data[train_indices], grouping_data[test_indices]
            else:
                train_groups, test_groups = None, None

            path = join(main_path, f"split_{split_num}")
            os.mkdir(path)

            log.info("             Fitting model and making predictions...")
            # Catch the ValueError associated with not being able to convert string to float
            #try:
                #print(train_X, train_y)


                # For Keras model, save model summary to main_path and plot training/validation vals vs. epochs
            if 'KerasRegressor' in str(model.__class__.__name__):
                with open(join(main_path, 'keras_model_summary.txt'), 'w') as f:
                    with redirect_stdout(f):
                        model.summary()
                history = model.fit(train_X, train_y)
                plot_helper.plot_keras_history(model_history=history,
                                                   savepath=join(path,'keras_model_accuracy.png'),
                                                   plot_type='accuracy')
                plot_helper.plot_keras_history(model_history=history,
                                                   savepath=join(path, 'keras_model_loss.png'),
                                                   plot_type='loss')
                pd.DataFrame().from_dict(data=history.history).to_excel(join(path,'keras_model_data.xlsx'))
            else:
                model.fit(train_X, train_y)

            #except ValueError:
            #    raise utils.InvalidValue('MAST-ML has detected an error with one of your feature vectors which has caused an error'
            #                       ' in model fitting.')
            # Save off the trained model as .pkl for future import

            # TODO: note that saving keras models has broken with updated keras version
            if 'KerasRegressor' not in model.__class__.__name__:
                joblib.dump(model, os.path.abspath(join(path, str(model.__class__.__name__)+"_split_"+str(split_num)+".pkl")))

            if is_classification:
                # For classification, need probabilty of prediction to make accurate ROC curve (and other predictions??).
                #TODO:Consider using only predict_proba and not predict() method for classif problems. Have exit escape if probability set to False here.
                # See stackoverflow post:
                #https: // stats.stackexchange.com / questions / 329857 / what - is -the - difference - between - decision
                # - function - predict - proba - and -predict - fun

                #params = model.get_params()
                #if params['probability'] == True:
                try:
                    train_pred_proba = model.predict_proba(train_X)
                    test_pred_proba = model.predict_proba(test_X)
                except:
                    log.error('You need to perform classification with model param probability=True enabled for accurate'
                                ' predictions, if your model has the probability param (e.g. RandomForestClassifier does not. '
                              'Please reset this parameter as applicable and re-run MASTML')
                    exit()
                train_pred = model.predict(train_X)
                test_pred = model.predict(test_X)
            else:
                train_pred = model.predict(train_X)
                test_pred  = model.predict(test_X)
                if train_pred.ndim > 1:
                    train_pred = np.squeeze(train_pred)
                if test_pred.ndim > 1:
                    test_pred = np.squeeze(test_pred)
                if conf['MiscSettings']['normalize_target_feature'] is True:
                    train_pred = normalizer_instance.inverse_transform(train_pred)
                    test_pred = normalizer_instance.inverse_transform(test_pred)
                    train_y = pd.Series(normalizer_instance.inverse_transform(train_y))
                    test_y = pd.Series(normalizer_instance.inverse_transform(test_y))

                # Here- for Random Forest, Extra Trees, and Gradient Boosters output feature importances
                if model.__class__.__name__ in ['RandomForestRegressor', 'ExtraTreesRegressor', 'GradientBoostingRegressor']:
                    pd.concat([pd.DataFrame(X.columns), pd.DataFrame(model.feature_importances_)],  1).to_excel(join(path, str(model.__class__.__name__)+'_featureimportances.xlsx'), index=False)

            # here is where we need to collect validation stats
            if is_validation:
                validation_predictions_list = list()
                validation_y_forpred_list = list()
                for validation_column_name in validation_column_names:
                    validation_X_forpred = _only_validation(X, validation_columns[validation_column_name])
                    validation_y_forpred = _only_validation(y, validation_columns[validation_column_name])
                    log.info("             Making predictions on prediction_only data...")
                    validation_predictions = model.predict(validation_X_forpred)
                    validation_predictions_list.append(validation_predictions)
                    validation_y_forpred_list.append(validation_y_forpred)

                    # save them as 'predicitons.csv'
                    validation_predictions = np.squeeze(validation_predictions)
                    validation_predictions_series = pd.Series(validation_predictions, name='clean_predictions', index=validation_X_forpred.index)
                    #validation_noinput_series = pd.Series(X_noinput.index, index=validation_X.index)
                    pd.concat([validation_X_forpred,  validation_y_forpred,  validation_predictions_series],  1)\
                            .to_csv(join(path, 'predictions_'+str(validation_column_name)+'.csv'), index=False)
            else:
                validation_y = None
            

            # Save train and test data and results to csv:
            log.info("             Saving train/test data and predictions to csv...")
            train_pred_series = pd.DataFrame(train_pred, columns=['train_pred'], index=train_indices)
            train_noinput_series = pd.DataFrame(X_noinput, index=train_indices)
            pd.concat([train_X, train_y, train_pred_series, train_noinput_series], 1)\
                    .to_csv(join(path, 'train.csv'), index=False)
            test_pred_series = pd.DataFrame(test_pred,   columns=['test_pred'],  index=test_indices)
            test_noinput_series = pd.DataFrame(X_noinput, index=test_indices)
            pd.concat([test_X,  test_y,  test_pred_series, test_noinput_series],  1)\
                    .to_csv(join(path, 'test.csv'),  index=False)


            log.info("             Calculating score metrics...")
            split_path = main_path.split(os.sep)

            # collect metrics inside a warning catching block for some things we know we should ignore
            with warnings.catch_warnings():
                # NOTE I tried making this more specific use warnings's regex filter but it would never
                # catch it for some indeterminiable reason.
                # This warning is raised when you ask for Recall on something from y_true that never
                # occors in y_pred. sklearn assumes 0.0, and we want it to do so (silently).
                warnings.simplefilter('ignore', UndefinedMetricWarning)
                train_metrics = OrderedDict((name, function(train_y, train_pred))
                                            for name, (_, function) in metrics_dict.items())
                test_metrics = OrderedDict((name, function(test_y, test_pred))
                                           for name, (_, function) in metrics_dict.items())
                # Need to pass y_train data to get rmse/sigma for test rmse and sigma of train y
                if 'rmse_over_stdev' in metrics_dict.keys():
                    test_metrics['rmse_over_stdev'] = metrics_dict['rmse_over_stdev'][1](test_y, test_pred, train_y)
                if 'R2_adjusted' in metrics_dict.keys():
                    test_metrics['R2_adjusted'] = metrics_dict['R2_adjusted'][1](test_y, test_pred, test_X.shape[1])
                    train_metrics['R2_adjusted'] = metrics_dict['R2_adjusted'][1](train_y, train_pred, train_X.shape[1])

                split_result = OrderedDict(
                    normalizer=split_path[-4],
                    selector=split_path[-3],
                    model=split_path[-2],
                    splitter=split_path[-1],
                    split_num=split_num,
                    y_train_true=train_y.values,
                    y_train_pred=train_pred,
                    y_test_true=test_y.values,
                    y_test_pred=test_pred,
                    train_metrics=train_metrics,
                    test_metrics=test_metrics,
                    train_indices=train_indices,
                    test_indices=test_indices,
                    train_groups=train_groups,
                    test_groups=test_groups,
                )

                if is_validation:
                    prediction_metrics_list = list()
                    for validation_column_name, validation_y, validation_predictions in zip(validation_column_names, validation_y_forpred_list, validation_predictions_list):
                        prediction_metrics = OrderedDict((name, function(validation_y, validation_predictions))
                                           for name, (_, function) in metrics_dict.items())
                        if 'rmse_over_stdev' in prediction_metrics.keys():
                            # Correct series passed?
                            prediction_metrics['rmse_over_stdev'] = metrics_dict['rmse_over_stdev'][1](validation_y, validation_predictions, train_y)
                        prediction_metrics_list.append(prediction_metrics)
                        split_result['y_validation_true'+'_'+str(validation_column_name)] = validation_y.values
                        split_result['y_validation_pred'+'_'+str(validation_column_name)] = validation_predictions
                    split_result['prediction_metrics'] = prediction_metrics_list
                else:
                    split_result['prediction_metrics'] = None

            if is_classification:
                split_result['y_train_pred_proba'] = train_pred_proba
                split_result['y_test_pred_proba'] = test_pred_proba

            log.info("             Making plots...")
            if MiscSettings['plot_train_test_plots']:
                plot_helper.make_train_test_plots(
                        split_result, path, is_classification, 
                        label=y.name, model=model, train_X=train_X, test_X=test_X, groups=grouping_data)

            if MiscSettings['plot_error_plots']:
                if is_validation:
                    plot_helper.make_error_plots(split_result, path, is_classification,
                                                 label=y.name, model=model, train_X=train_X, test_X=test_X,
                                                 rf_error_method=MiscSettings['rf_error_method'],
                                                 rf_error_percentile=MiscSettings['rf_error_percentile'],
                                                 is_validation = is_validation,
                                                 validation_column_name = validation_column_name,
                                                 validation_X = validation_X_forpred,
                                                 groups=grouping_data)
                else:
                    plot_helper.make_error_plots(split_result, path, is_classification,
                                                 label=y.name, model=model, train_X=train_X, test_X=test_X,
                                                 rf_error_method=MiscSettings['rf_error_method'],
                                                 rf_error_percentile=MiscSettings['rf_error_percentile'],
                                                 is_validation = is_validation,
                                                 validation_column_name = None,
                                                 validation_X= None,
                                                 groups=grouping_data)
            # Write stats in each split path, not main path
            if is_validation:
                _write_stats_tocsv(split_result['train_metrics'],
                         split_result['test_metrics'],
                         path,
                         split_result['prediction_metrics'],
                         validation_column_names)
                _write_stats(split_result['train_metrics'],
                         split_result['test_metrics'],
                         path,
                         split_result['prediction_metrics'],
                         validation_column_names)
            else:
                _write_stats_tocsv(split_result['train_metrics'],
                             split_result['test_metrics'],
                             path)
                _write_stats(split_result['train_metrics'],
                             split_result['test_metrics'],
                             path)

            return split_result

        split_results = []
        for split_num, (train_indices, test_indices) in enumerate(trains_tests):
            split_results.append(one_fit(split_num, train_indices, test_indices, normalizer_instance))

        log.info("    Calculating mean and stdev of scores...")
        def make_train_test_average_and_std_stats():
            train_stats = OrderedDict([('Average Train', None)])
            test_stats  = OrderedDict([('Average Test', None)])
            if is_validation:
                prediction_stats = list()
                num_predictions = len(split_results[0]['prediction_metrics'])
                for i in range(num_predictions):
                    prediction_stats.append(OrderedDict([('Average Prediction', None)]))
            for name in metrics_dict:
                train_values = [split_result['train_metrics'][name] for split_result in split_results]
                test_values  = [split_result['test_metrics'][name]  for split_result in split_results]
                train_stats[name] = (np.mean(train_values), np.std(train_values))
                test_stats[name]  = (np.mean(test_values), np.std(test_values))
                if is_validation:
                    for i in range(num_predictions):
                        prediction_values = [split_result['prediction_metrics'][i][name] for split_result in split_results]
                        prediction_stats[i][name] = (np.mean(prediction_values), np.std(prediction_values))
                test_stats_single = dict()
                test_stats_single[name] = (np.mean(test_values), np.std(test_values))
                if grouping_data is not None:
                    groups = np.array(split_results[0]['test_groups'].tolist()+split_results[0]['train_groups'].tolist())
                    unique_groups = np.union1d(split_results[0]['test_groups'], split_results[0]['train_groups'])
                    plot_helper.plot_metric_vs_group(metric=name, groups=unique_groups, stats=test_values,
                                                     avg_stats = test_stats_single, savepath=join(main_path, str(name)+'_vs_group.png'))
                    plot_helper.plot_metric_vs_group_size(metric=name, groups=groups, stats=test_values,
                                                     avg_stats = test_stats_single, savepath=join(main_path, str(name)+'_vs_group_size.png'))
            del train_stats['Average Train']
            del test_stats['Average Test']
            if is_validation:
                for i in range(num_predictions):
                    del prediction_stats[i]['Average Prediction']
                return train_stats, test_stats, prediction_stats
            else:
                return train_stats, test_stats

        if is_validation:
            avg_train_stats, avg_test_stats, avg_prediction_stats = make_train_test_average_and_std_stats()
            # Here- write average stats to main folder of splitter
            _write_stats_tocsv(avg_train_stats,
                         avg_test_stats,
                         main_path, prediction_metrics=avg_prediction_stats, prediction_names=validation_column_names)
            _write_stats(avg_train_stats,
                         avg_test_stats,
                         main_path, prediction_metrics=avg_prediction_stats, prediction_names=validation_column_names)
        else:
            avg_train_stats, avg_test_stats = make_train_test_average_and_std_stats()
            # Here- write average stats to main folder of splitter
            _write_stats_tocsv(avg_train_stats,
                         avg_test_stats,
                         main_path)
            _write_stats(avg_train_stats,
                         avg_test_stats,
                         main_path)

        def make_average_error_plots(main_path):
            has_model_errors = False
            has_model_errors_validation = False
            dfs_cumulative_errors = list()
            dfs_cumulative_errors_validation = list()
            for split_folder, _, __ in os.walk(main_path):
                if "split" in split_folder:
                    path = join(main_path, split_folder)
                    try:
                        dfs_cumulative_errors.append(pd.read_csv(join(path,'test_cumulative_normalized_error.csv')))
                        if is_validation:
                            dfs_cumulative_errors_validation.append(pd.read_csv(join(path, 'validation_cumulative_normalized_error.csv')))
                    except:
                        pass

            # Concatenate all dfs in list to one big df
            df_cumulative_errors = pd.concat(dfs_cumulative_errors)
            if is_validation:
                df_cumulative_errors_validation = pd.concat(dfs_cumulative_errors_validation)
            # Need to get average values of df columns by averagin over groups of Y True values (since each Y True should
            # only appear once)
            # TODO: change this to get values explicitly from each split and then average, as some Y True values may have same value and appear multiple times
            df_normalized_errors_avgvalues = df_cumulative_errors.groupby('Y True').mean().reset_index()
            y_true = np.array(df_normalized_errors_avgvalues['Y True'])
            y_pred = np.array(df_normalized_errors_avgvalues['Y Pred'])
            if is_validation:
                df_normalized_errors_avgvalues_validation = df_cumulative_errors_validation.groupby('Y True').mean().reset_index()
                y_true_validation = np.array(df_normalized_errors_avgvalues_validation['Y True'])
                y_pred_validation = np.array(df_normalized_errors_avgvalues_validation['Y Pred'])
            try:
                average_error_values = np.array(df_normalized_errors_avgvalues['error_bars_down'])
                has_model_errors = True
            except:
                average_error_values = None
                has_model_errors = False

            if is_validation:
                try:
                    average_error_values_validation = np.array(df_normalized_errors_avgvalues_validation['error_bars_down'])
                    has_model_errors_validation = True
                except:
                    average_error_values_validation = None
                    has_model_errors_validation = False

            plot_helper.plot_average_cumulative_normalized_error(y_true=y_true, y_pred=y_pred,
                                                                 savepath=join(main_path,'test_cumulative_normalized_error_average_allsplits.png'),
                                                                 has_model_errors=has_model_errors,
                                                                 err_avg=average_error_values)
            if is_validation:
                plot_helper.plot_average_cumulative_normalized_error(y_true=y_true_validation, y_pred=y_pred_validation,
                                                                     savepath=join(main_path,
                                                                                   'validation_cumulative_normalized_error_average_allsplits.png'),
                                                                     has_model_errors=has_model_errors_validation,
                                                                     err_avg=average_error_values_validation)
            plot_helper.plot_average_normalized_error(y_true=y_true, y_pred=y_pred,
                                                      savepath=join(main_path,'test_normalized_error_average_allsplits.png'),
                                                                 has_model_errors=has_model_errors,
                                                                 err_avg=average_error_values)
            return

        # Call to make average error plots
        if conf['MiscSettings']['plot_error_plots']:
            log.info("    Making average error plots over all splits")
            make_average_error_plots(main_path=main_path)

        log.info("    Making best/worst plots...")
        def get_best_worst_median_runs():
            # sort splits by the test score of first metric:
            greater_is_better, _ = next(iter(metrics_dict.values())) # get first value pair
            scalar = 1 if greater_is_better else -1
            s = sorted(split_results, key=lambda run: scalar*next(iter(run['test_metrics'])))
            return s[0], s[len(split_results)//2], s[-1]
        worst, median, best = get_best_worst_median_runs()

        def make_pred_vs_true_plots(model, y):
            if conf['MiscSettings']['normalize_target_feature'] == True:
                y = pd.Series(normalizer_instance.inverse_transform(y), name=conf['GeneralSetup']['input_target'])

            if MiscSettings['plot_predicted_vs_true']:
                plot_helper.plot_best_worst_split(y.values, best, worst,
                                                  join(main_path, 'best_worst_split'), label=conf['GeneralSetup']['input_target'])
            predictions = [[] for _ in range(X.shape[0])]
            for split_num, (train_indices, test_indices) in enumerate(trains_tests):
                for i, pred in zip(test_indices, split_results[split_num]['y_test_pred']):
                    predictions[i].append(pred)
            if MiscSettings['plot_predicted_vs_true_average']:
                plot_helper.plot_predicted_vs_true_bars(
                        y.values, predictions, avg_test_stats,
                        join(main_path, 'predicted_vs_true_average'), label=conf['GeneralSetup']['input_target'])
                if grouping_data is not None:
                    plot_helper.plot_predicted_vs_true_bars(
                        y.values, predictions, avg_test_stats,
                        join(main_path, 'predicted_vs_true_average_groupslabeled'),
                        label=conf['GeneralSetup']['input_target'],
                        groups=grouping_data)
            if MiscSettings['plot_best_worst_per_point']:
                plot_helper.plot_best_worst_per_point(y.values, predictions,
                                                      join(main_path, 'best_worst_per_point'),
                                                      metrics_dict, avg_test_stats, label=conf['GeneralSetup']['input_target'])

        if not is_classification:
            make_pred_vs_true_plots(model=model, y=y)

        return split_results

    runs = do_all_combos(X, y, df) # calls do_one_splitter internally

    log.info("Making image html file...")
    html_helper.make_html(outdir)

    log.info("Making html file of all runs stats...")
    _save_all_runs(runs, outdir)

    # Here- do DLHub model hosting if have section
    if bool(conf['ModelHosting']) != False: # dict is empty
        model_hosting.host_model(model_path=conf['ModelHosting']['model_path'],
                                 preprocessor_path=conf['ModelHosting']['preprocessor_path'],
                                 training_data_path=conf['ModelHosting']['training_data_path'],
                                 model_title=conf['ModelHosting']['model_title'],
                                 model_name=conf['ModelHosting']['model_name'],
                                 model_type="scikit-learn")
        log.info('Finished uploading model to DLHub...')

    log.info('Your MAST-ML run has finished successfully!')
    return

def _instantiate(kwargs_dict, name_to_constructor, category, X_grouped=None, X_indices=None):
    """
    Uses name_to_constructor to instantiate every item in kwargs_dict and return
    the list of instantiations
    """
    instantiations = []
    for long_name, (name, kwargs) in kwargs_dict.items():
        log.debug(f'instantiation: {long_name}, {name}({kwargs})')
        try:
            #skip instantiate step for keras model because need to pass dict to build model and not all values directly
            if 'KerasRegressor' in long_name:
                pass

            # Need to construct cv object when have special case of RFECV and LeaveOneGroupOut cross-validation!
            elif name == 'RFECV':
                if 'cv' in kwargs.keys():
                    if X_grouped is not None:
                        if kwargs['cv'].__class__.__name__ == 'LeaveOneGroupOut':
                            trains = list()
                            tests = list()
                            for train_idx, test_idx in LeaveOneGroupOut().split(X=X_indices, y=None, groups=X_grouped):
                                trains.append(train_idx)
                                tests.append(test_idx)
                            custom_cv = zip(trains, tests)
                            kwargs['cv'] = custom_cv
                instantiations.append([long_name, name_to_constructor[name](**kwargs)])
            else:
                instantiations.append([long_name, name_to_constructor[name](**kwargs)])

        except TypeError:
            log.info(f"ARGUMENTS FOR '{name}': {inspect.signature(name_to_constructor[name])}")
            raise utils.InvalidConfParameters(
                f"The {category} '{name}' has invalid parameters: {kwargs}\n"
                f"Signature for '{name}': {inspect.signature(name_to_constructor[name])}")
        except KeyError:
            raise utils.InvalidConfSubSection(
                f"There is no {category} called '{name}'."
                f"All valid {category}: {list(name_to_constructor.keys())}")

    return instantiations

def _grouping_column_to_group_number(X_grouped):
    group_list = X_grouped.values.reshape((1, -1))
    unique_groups = np.unique(group_list).tolist()
    group_dict = dict()
    group_list_asnumber = list()
    for i, group in enumerate(unique_groups):
        group_dict[group] = i+1
    for i, group in enumerate(group_list.tolist()[0]):
        group_list_asnumber.append(group_dict[group])
    X_grouped_asnumber = np.asarray(group_list_asnumber)
    return X_grouped_asnumber

def _snatch_models(models, conf_feature_selection):
    models = OrderedDict(models)
    log.debug(f'models, pre-snatching: \n{models}')
    for selector_name, [_, args_dict] in conf_feature_selection.items():
        if 'estimator' in args_dict:
            model_name = args_dict['estimator']
            try:
                args_dict['estimator'] = models[model_name]
                del models[model_name]
            except KeyError:
                raise utils.MastError(f"The selector {selector_name} specified model {model_name},"
                                      f"which was not found in the [Models] section")
    log.debug(f'models, post-snatching: \n{models}')
    return models

def _snatch_keras_model(models, conf_models):
    for model in conf_models.keys():
        if 'KerasRegressor' in model:
            keras_model = model_finder.KerasRegressor(conf_models[model][1])
            models[model] = keras_model
    return models

def _snatch_gpr_model(models, conf_models):
    for model in models.keys():
        if 'GaussianProcessRegressor' in model:
            import sklearn.gaussian_process
            from sklearn.gaussian_process import GaussianProcessRegressor
            kernel_list = ['WhiteKernel', 'RBF', 'ConstantKernel', 'Matern', 'RationalQuadratic', 'ExpSineSquared', 'DotProduct']
            kernel_operators = ['+', '*', '-']
            params = conf_models[model]
            kernel_string = params[1]['kernel']
            # Need to delete old kernel (as str) from params so can use other specified params in new GPR model
            del params[1]['kernel']
            # Parse kernel_string to identify kernel types and any kernel operations to combine kernels
            kernel_types_asstr = list()
            kernel_types_ascls = list()
            kernel_operators_used = list()

            for s in kernel_string[:]:
                if s in kernel_operators:
                    kernel_operators_used.append(s)

            # Do case for single kernel, no operators
            if len(kernel_operators_used) == 0:
                kernel_types_asstr.append(kernel_string)
            else:
                # New method, using re
                unique_operators = np.unique(kernel_operators_used).tolist()
                unique_operators_asstr = '['
                for i in unique_operators:
                    unique_operators_asstr += str(i)
                unique_operators_asstr += ']'
                kernel_types_asstr = re.split(unique_operators_asstr, kernel_string)

            for kernel in kernel_types_asstr:
                kernel_ = getattr(sklearn.gaussian_process.kernels, kernel)
                kernel_types_ascls.append(kernel_())

            # Case for single kernel
            if len(kernel_types_ascls) == 1:
                kernel = kernel_types_ascls[0]

            kernel_count = 0
            for i, operator in enumerate(kernel_operators_used):
                if i+1 <= len(kernel_operators_used):
                    if operator == "+":
                        if kernel_count == 0:
                            kernel = kernel_types_ascls[kernel_count] + kernel_types_ascls[kernel_count+1]
                        else:
                            kernel += kernel_types_ascls[kernel_count+1]
                    elif operator == "*":
                        if kernel_count == 0:
                            kernel = kernel_types_ascls[kernel_count] * kernel_types_ascls[kernel_count+1]
                        else:
                            kernel *= kernel_types_ascls[kernel_count+1]
                    else:
                        logging.warning('You have chosen an invalid operator to construct a composite kernel. Please choose'
                                      ' either "+" or "*".')
                    kernel_count += 1

            gpr = GaussianProcessRegressor(kernel=kernel, **params[1])
            # Need to delete old GPR from model list and replace with new GPR with correct kernel and other params.
            del models[model]
            models[model] = gpr
            break
    return models

def _snatch_models_cv_for_hyperopt(conf, models, splitters):
    models = list(models.items())
    if conf['HyperOpt']:
        for searchtype, searchparams in conf['HyperOpt'].items():
            for paramtype, paramvalue in searchparams[1].items():
                if paramtype == 'estimator':
                    # Need to grab model and params from Model section of conf file
                    found_model = False
                    for model in models:
                        if model[0] == paramvalue:
                            conf['HyperOpt'][searchtype][1]['estimator'] = model[1]
                            found_model = True
                            break
                    if found_model == False:
                        raise utils.MastError(f"The estimator {paramvalue} could not be found in the input file!")
                if paramtype == 'cv':
                    # Need to grab cv and params from DataSplits section of conf file
                    found_cv = False
                    for splitter in splitters:
                        if splitter[0] == paramvalue:
                            conf['HyperOpt'][searchtype][1]['cv'] = splitter[1]
                            found_cv = True
                            break
                    if found_cv == False:
                        raise utils.MastError(f"The cv object {paramvalue} could not be found in the input file!")
                if paramtype == 'scoring':
                    # Need to grab correct scoring object
                    found_scorer = False
                    metrics_dict = metrics.regression_metrics
                    if paramvalue in metrics_dict.keys():
                        conf['HyperOpt'][searchtype][1]['scoring'] = make_scorer(metrics_dict[paramvalue][1],
                                                                                 greater_is_better=metrics_dict[paramvalue][0])
                        found_scorer = True
                        break
                    if found_scorer == False:
                        raise utils.MastError(
                            f"The scoring object {paramvalue} could not be found in the input file!")

    return conf['HyperOpt']

def _snatch_splitters(splitters, conf_feature_selection):
    log.debug(f'cv, pre-snatching: \n{splitters}')
    for selector_name, (_, args_dict) in conf_feature_selection.items():
        # Here: add snatch to cv object for feature selection with RFECV
        if 'cv' in args_dict:
            cv_name = args_dict['cv']
            try:
                args_dict['cv'] = splitters[cv_name]
                del splitters[cv_name]
            except KeyError:
                raise utils.MastError(f"The selector {selector_name} specified cv splitter {cv_name},"
                                      f"which was not found in the [DataSplits] section")
    log.debug(f'cv, post-snatching: \n{splitters}')

def _extract_grouping_column_names(splitter_to_kwargs):
    splitter_to_group_names = dict()
    for splitter_name, name_and_kwargs in splitter_to_kwargs.items():
        _, kwargs = name_and_kwargs
        if 'grouping_column' in kwargs:
            column_name = kwargs['grouping_column']
            del kwargs['grouping_column'] # because the splitter doesn't actually take this
            splitter_to_group_names[splitter_name] = column_name
    return splitter_to_group_names

def _remove_constant_features(df):
    log.info("Removing constant features, regardless of feature selectors.")
    before = set(df.columns)
    df = df.loc[:, (df != df.iloc[0]).any()]
    removed = list(before - set(df.columns))
    if removed != []:
        log.warning(f'Removed {len(removed)}/{len(before)} constant columns.')
        log.debug("Removed the following constant columns: " + str(removed))
    return df

def _save_all_runs(runs, outdir):
    """
    Produces a giant html table of all stats for all runs
    """
    table = []
    for run in runs:
        od = OrderedDict()
        for name, value in run.items():
            if name == 'train_metrics':
                for k, v in run['train_metrics'].items():
                    od['train_'+k] = v
            elif name == 'test_metrics':
                for k, v in run['test_metrics'].items():
                    od['test_'+k] = v
            else:
                od[name] = value
        table.append(od)
    pd.DataFrame(table).to_html(join(outdir, 'all_runs_table.html'))

def _write_stats(train_metrics, test_metrics, outdir, prediction_metrics=None, prediction_names=None):
    with open(join(outdir, 'stats_summary.txt'), 'w') as f:
        f.write("TRAIN:\n")
        for name,score in train_metrics.items():
            if type(score) == tuple:
                f.write(f"{name}: {'%.3f'%float(score[0])} +/- {'%.3f'%float(score[1])}\n")
            else:
                f.write(f"{name}: {'%.3f'%float(score)}\n")
        f.write("TEST:\n")
        for name,score in test_metrics.items():
            if type(score) == tuple:
                f.write(f"{name}: {'%.3f'%float(score[0])} +/- {'%.3f'%float(score[1])}\n")
            else:
                f.write(f"{name}: {'%.3f'%float(score)}\n")
        if prediction_metrics:
            #prediction metrics now list of dicts for predicting multiple values
            for prediction_metric, prediction_name in zip(prediction_metrics, prediction_names):
                f.write("PREDICTION for "+str(prediction_name)+":\n")
                for name, score in prediction_metric.items():
                    if type(score) == tuple:
                        f.write(f"{name}: {'%.3f'%float(score[0])} +/- {'%.3f'%float(score[1])}\n")
                    else:
                        f.write(f"{name}: {'%.3f'%float(score)}\n")

def _write_stats_tocsv(train_metrics, test_metrics, outdir, prediction_metrics=None, prediction_names=None):
    datadict = dict()
    for name, score in train_metrics.items():
        if type(score) == tuple:
            datadict[name+' train score'] = float(score[0])
            datadict[name+' train stdev'] = float(score[1])
        else:
            datadict[name+' train score'] = float(score)
    for name, score in test_metrics.items():
        if type(score) == tuple:
            datadict[name+' validation score'] = float(score[0])
            datadict[name+' validation stdev'] = float(score[1])
        else:
            datadict[name+' validation score'] = float(score)
    if prediction_names:
        for prediction_metric, prediction_name in zip(prediction_metrics, prediction_names):
            for name, score in prediction_metric.items():
                if type(score) == tuple:
                    datadict[prediction_name+' '+name+' score'] = float(score[0])
                    datadict[prediction_name+' '+name+' stdev'] = float(score[1])
                else:
                    datadict[prediction_name+' '+name+' score'] = float(score)
    pd.DataFrame().from_dict(data=datadict, orient='index').to_csv(join(outdir, 'stats_summary.csv'))
    return

def _exclude_validation(df, validation_column):
    return df.loc[validation_column != 1]

def _only_validation(df, validation_column):
    return df.loc[validation_column == 1]

def check_paths(conf_path, data_path, outdir):
    """
    This method is responsible for error handling of the user-specified paths for the configuration file, data file,
    and output directory.

    Args:

        conf_path: (str), the path supplied by the user which contains the input configuration file

        data_path: (str), the path supplied by the user which contains the input data file (as CSV or XLSX)

        outdir: (str), the path supplied by the user which determines where the output results are saved to

    Returns:

        conf_path: (str), the path supplied by the user which contains the input configuration file

        data_path: (str), the path supplied by the user which contains the input data file (as CSV or XLSX)

        outdir: (str), the path supplied by the user which determines where the output results are saved to

    """


    # Check conf path:
    if type(conf_path) is str:
        if os.path.splitext(conf_path)[1] != '.conf':
            raise utils.FiletypeError(f"Conf file does not end in .conf: '{conf_path}'")
        if not os.path.isfile(conf_path):
            raise utils.FileNotFoundError(f"No such file: {conf_path}")
    elif type(conf_path) is dict:
        pass
    else:
        raise TypeError('Your conf_path must be either a string to .conf file path or a dict')

    # Check data path:
    if type(data_path) is str:
        if os.path.splitext(data_path)[1] not in ['.csv', '.xlsx']:
            raise utils.FiletypeError(f"Data file does not end in .csv or .xlsx: '{data_path}'")
        if not os.path.isfile(data_path):
            raise utils.FileNotFoundError(f"No such file: {data_path}")
    elif type(data_path) is type(pd.DataFrame()):
        pass
    else:
        raise TypeError('Your data_path must be either a string to .csv or .xlsx data file or a pd.DataFrame object')

    # Check output directory:

    if os.path.exists(outdir):
        try:
            os.rmdir(outdir) # succeeds if empty
        except OSError: # directory not empty
            log.warning(f"{outdir} not empty. Renaming...")
            now = datetime.now()
            outdir = outdir.rstrip(os.sep) # remove trailing slash
            outdir = f"{outdir}_{now.month:02d}_{now.day:02d}" \
                     f"_{now.hour:02d}_{now.minute:02d}_{now.second:02d}"
    os.makedirs(outdir)
    log.info(f"Saving to directory '{outdir}'")

    return conf_path, data_path, outdir

def get_commandline_args():
    """
    This method is responsible for parsing and checking the command-line execution of MAST-ML inputted by the user.

    Args:

        None

    Returns:

        (str), the path supplied by the user which contains the input configuration file

        (str), the path supplied by the user which contains the input data file (as CSV or XLSX)

        (str), the path supplied by the user which determines where the output results are saved to

        verbosity: (int), the verbosity level of the MAST-ML log, which determines the amount of information writtent to the log.

    """

    parser = argparse.ArgumentParser(description='MAterials Science Toolkit - Machine Learning')
    parser.add_argument('conf_path', type=str, help='path to mastml .conf file')
    parser.add_argument('data_path', type=str, help='path to csv or xlsx file')
    parser.add_argument('-o', action="store", dest='outdir', default='results',
                        help='Folder path to save output files to. Defaults to results/')
    # from https://stackoverflow.com/a/14763540
    # we only use them to set a bool but it would be nice to have multiple levels in the future
    parser.add_argument('-v', '--verbosity', action="count",
                        help="include this flag for more verbose output")
    parser.add_argument('-q', '--quietness', action="count",
                       help="include this flag to hide [DEBUG] printouts, or twice to hide [INFO]")

    args = parser.parse_args()
    verbosity = (args.verbosity if args.verbosity else 0)\
            - (args.quietness if args.quietness else 0)
    # verbosity -= 1 ## uncomment this for distribution
    return (os.path.abspath(args.conf_path),
            os.path.abspath(args.data_path),
            os.path.abspath(args.outdir),
            verbosity)

if __name__ == '__main__':
    conf_path, data_path, outdir, verbosity = get_commandline_args()
    main(conf_path, data_path, outdir, verbosity)
